import torch


########################################################################################################################
######################################### Modules & Helper Functions ###################################################
########################################################################################################################
def time_conv_Allx1(input_units):
    """Allx1 convolution over time (taking all input channels into account at each timestep"""
    return torch.nn.Conv1d(input_units, input_units, kernel_size=1, padding=0, groups=1)


def time_conv_1x3(input_units, depth="channelwise"):
    """1x3 convolution over time channelwise """
    if depth == "channelwise":
        groups = input_units
    elif depth == "full":
        groups = 1
    else:
        groups = depth

    return torch.nn.Conv1d(input_units, input_units, kernel_size=3, padding=1, groups=groups)


def time_conv_1x5(input_units, depth="channelwise"):
    """1x5 convolution over time channelwise """
    assert depth in ["channelwise", "full"], "depth can only be channelwise or full"
    if depth == "channelwise":
        groups = input_units
    elif depth == "full":
        groups = 1
    else:
        groups = depth
    return torch.nn.Conv1d(input_units, input_units, kernel_size=5, padding=2, groups=groups)

def add_vel_and_acc_info(x):
    """
     Approximate Velocity and Accelartion and add to feature dimension
    :param x: torch.Tensor
        3D-Tensor (batch, seq_length, channels)
    :return: torch.Tensor
        3D-Tensor (batch,seq_length, 3*channels)
    """
    zeros = torch.zeros(x.shape[0], 1, x.shape[2],device = x.device)#.to(DEVICE)
    velocity = x[:, 1:, :] - x[:, :-1, :]
    acceleration = velocity[:, 1:, :] - velocity[:, :-1, :]
    velocity = torch.cat((velocity, zeros), axis=1)
    acceleration = torch.cat((zeros, acceleration, zeros), axis=1)
    x = torch.cat((x, velocity, acceleration), axis=2)
    return x

def double_sequence(x):
    """
    Interpolate between time steps
    :param x: torch.Tensor
        3D-Tensor (batch, seq_length, channels)
    :return: torch.Tensor
        3D-Tensor (batch, 2*seq_length, channels)
    """
    x1 = x
    x2 = (x[:, :-1, :] + x[:, 1:, :]) / 2.0
    x2 = torch.cat([x2, x1[:, -1, :].view(x1.shape[0], 1, x1.shape[2])], axis=1)

    #x = torch.zeros((x1.shape[0], x1.shape[1] + x2.shape[1], x1.shape[2]), dtype=torch.double, requires_grad=True, device=DEVICE)
    x = torch.zeros((x1.shape[0], x1.shape[1] + x2.shape[1], x1.shape[2]), dtype=torch.double,
                    requires_grad=False, device = x.device)
    x[:, ::2, :] = x1  # Index every second row, starting from 0
    x[:, 1::2, :] = x2

    return x

class TimeConvIncpetionBlock(torch.nn.Module):
    def __init__(self,input_units, pre_activation_function, add_resid=True):
        super(TimeConvIncpetionBlock, self).__init__()

        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.band_conv1d_1 = time_conv_Allx1(input_units)
        self.band_conv1d_3 = time_conv_1x3(input_units)
        self.band_conv1d_5 = time_conv_1x5(input_units)
        self.band_conv1d_combine = torch.nn.Conv1d(3 * input_units, input_units, 1, padding=0, groups=input_units)

        self.activation = pre_activation_function
        self.add_resid = add_resid

    def forward(self, x):
        resid = x

        out = self.activation(x)
        out1 = self.band_conv1d_1(out)
        out3 = self.band_conv1d_3(out)
        out5 = self.band_conv1d_5(out)


        out = [torch.stack((out1[:, i, :], out3[:, i, :], out5[:, i, :]), axis=1) for i in range(out.shape[1])]
        out = torch.cat(out, axis=1)
        if self.add_resid:
            out += resid

        return out



class TimeConvResBlock(torch.nn.Module):
    def __init__(self,input_units,filter_size,pre_activation_function = torch.nn.Identity(),
                 post_activation_function = torch.nn.Identity(),
                 add_resid=True, depth="channelwise"):
        super(TimeConvResBlock, self).__init__()
        assert filter_size in [5, 3], "should be a valid filter size (5 or 3)"
        if filter_size == 3:
            self.band_conv1d_1 = time_conv_1x3(input_units,depth)
            self.band_conv1d_2 = time_conv_1x3(input_units,depth)
        else:
            self.band_conv1d_1 = time_conv_1x5(input_units,depth)
            self.band_conv1d_2 = time_conv_1x5(input_units,depth)

        self.pre_activation = pre_activation_function
        self.post_activation = post_activation_function
        self.add_resid = add_resid

    def forward(self,x):
        resid = x

        out = self.band_conv1d_1(self.pre_activation(x))
        out = self.band_conv1d_2(out)
        out = self.post_activation(out)
        if self.add_resid:
            out += resid
        return out


class MelChannelConv1D(torch.nn.Module):
    def __init__(self, input_units, filter_size_channel):
        super(MelChannelConv1D, self).__init__()
        self.filter_size_channel = filter_size_channel
        assert input_units % filter_size_channel == 0, 'output_size has to devisible by %d' % filter_size_channel
        output_units = int(input_units // filter_size_channel)
        self.ConvLayers = torch.nn.ModuleList(
            [torch.nn.Conv1d(input_units, output_units, 5, padding=2, groups=output_units) for _ in
             range(filter_size_channel)])

    def forward(self, x):
        batch, mel, seq = x.shape
        xs = []
        for i in range(self.filter_size_channel - 2):
            x_i = torch.cat((torch.zeros(batch, i + 1, seq, device = x.device), x[:, :-(i + 1), :]), axis=1)
            xs.append(x_i)
        xs.append(x)
        xs.append(torch.cat((x[:, 1:, :], torch.zeros(batch, 1, seq, device = x.device)), axis=1))

        outputs = []
        for i, layer in enumerate(self.ConvLayers):
            output_i = layer(xs[i])
            outputs.append(output_i)

        output = [torch.stack([res[:, i, :] for res in outputs], axis=1) for i in range(output_i.shape[1])]
        output = torch.cat(output, axis=1)

        return output



########################################################################################################################
################################################# Inverse Models  ######################################################
########################################################################################################################

class InverseModelMelTimeSmoothResidual(torch.nn.Module):
    """
        InverseModel
        - Initial Conv1d Layers for Convolution over time and neighbouring Mel Channels with residual connections
        - stacked LSTM-Cells
        - Post Conv1d Layers for Convolution over time stacked with residual connections
        - Weighting of lstm output and smoothed output
    """

    def __init__(self, input_size=60,
                 output_size=30,
                 hidden_size=180,
                 num_lstm_layers=4,
                 mel_smooth_layers=3,
                 mel_smooth_filter_size=3,
                 mel_resid_activation = torch.nn.Identity(),
                 resid_blocks=5,
                 time_filter_size=5,
                 pre_resid_activation = torch.nn.Identity(),
                 post_resid_activation = torch.nn.Identity(),
                 output_activation = torch.nn.Identity(),
                 lstm_resid=True):
        super().__init__()

        self.lstm_resid = lstm_resid
        self.mel_resid_activation = mel_resid_activation
        self.pre_activation = pre_resid_activation
        self.post_activation = post_resid_activation
        self.output_activation = output_activation

        self.double_sequence = double_sequence
        self.add_vel_and_acc_info = add_vel_and_acc_info
        self.MelBlocks = torch.nn.ModuleList(
            [MelChannelConv1D(input_size, mel_smooth_filter_size) for _ in range(mel_smooth_layers)])
        self.lstm = torch.nn.LSTM(3 * (input_size), hidden_size, num_layers=num_lstm_layers, batch_first=True)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)
        self.ResidualConvBlocks = torch.nn.ModuleList(
            [TimeConvResBlock(output_size, time_filter_size, self.pre_activation, self.post_activation) for _ in
             range(resid_blocks)])

        if self.lstm_resid and len(self.ResidualConvBlocks) > 0:
            self.resid_weighting = torch.nn.Conv1d(2 * output_size, output_size, time_filter_size, padding=2,
                                               groups=output_size)

    def forward(self, x, *args):
        if len(self.MelBlocks) > 0:
            x = x.permute(0, 2, 1)
            for layer in self.MelBlocks:
                shortcut = x
                x = layer(x)
                x += shortcut
                x = self.mel_resid_activation(x)
            x = x.permute(0, 2, 1)

        x = self.add_vel_and_acc_info(x)
        output, _ = self.lstm(x)
        output = self.post_linear(output)
        output = self.double_sequence(output)

        output = output.permute(0, 2, 1)
        lstm_output = output
        for layer in self.ResidualConvBlocks:
            output = layer(output)

        if len(self.ResidualConvBlocks) > 0 and self.lstm_resid:
            output = [torch.stack((output[:, i, :], lstm_output[:, i, :]), axis=1) for i in range(output.shape[1])]
            output = torch.cat(output, axis=1)
            output = self.resid_weighting(output)

        output = self.output_activation(output.permute(0, 2, 1))
        return output




########################################################################################################################
################################################# Forward Models  ######################################################
########################################################################################################################

class ForwardModelMelTimeSmoothResidual(torch.nn.Module):
    """
        ForwardModel
        - Initial Conv1d layers for Convolution over time stacked with residual connections
        - stacked LSTM-Cells
        - Post Conv1d Layers for Convolution over time and neighbouring Mel Channels with residual connections
        - Weighting of lstm output and smoothed output
    """

    def __init__(self, input_size=30,
                 output_size=60,
                 hidden_size=180,
                 num_lstm_layers=4,
                 mel_smooth_layers=3,
                 mel_smooth_filter_size=3,
                 mel_resid_activation = torch.nn.Identity(),
                 resid_blocks=5,
                 pre_resid_activation=torch.nn.Identity(),
                 post_resid_activation=torch.nn.Identity(),
                 time_filter_size=5,
                 lstm_resid = True,
                 output_activation = torch.nn.Identity()):
        super().__init__()

        self.lstm_resid = lstm_resid
        self.pre_activation = pre_resid_activation
        self.post_activation = post_resid_activation
        self.output_activation = output_activation
        self.mel_resid_activation = mel_resid_activation
        self.ResidualConvBlocks = torch.nn.ModuleList(
            [TimeConvResBlock(input_size, time_filter_size, self.pre_activation, self.post_activation) for _ in range(resid_blocks)])
        self.half_sequence = torch.nn.AvgPool1d(2, stride=2)
        self.add_vel_and_acc_info = add_vel_and_acc_info
        self.lstm = torch.nn.LSTM(3 * (input_size), hidden_size, num_layers=num_lstm_layers, batch_first = True)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)
        self.MelBlocks = torch.nn.ModuleList(
            [MelChannelConv1D(output_size, mel_smooth_filter_size) for _ in range(mel_smooth_layers)])

        if self.lstm_resid and len(self.MelBlocks) >0:
            self.resid_weighting = torch.nn.Conv1d(2 * output_size, output_size, time_filter_size, padding=2,
                                                   groups=output_size)

    def forward(self, x,*args):
        if len(self.ResidualConvBlocks) >0:
            x = x.permute(0, 2, 1)
            for layer in self.ResidualConvBlocks:
                x = layer(x)
            x = x.permute(0, 2, 1)
        x = self.add_vel_and_acc_info(x)
        output, _ = self.lstm(x)
        output = self.post_linear(output)
        output = output.permute(0, 2, 1)
        output = self.half_sequence(output)
        lstm_output = output

        for layer in self.MelBlocks:
            shortcut = output
            output = layer(output)
            output += shortcut
            output = self.mel_resid_activation(output)

        if len(self.MelBlocks) >0 and self.lstm_resid:
            output = [torch.stack((lstm_output[:, i, :], output[:, i, :]), axis=1) for i in range(output.shape[1])]
            output = torch.cat(output, axis=1)
            output = self.resid_weighting(output)
        output = output.permute(0, 2, 1)
        output = self.output_activation(output)

        return output

class ForwardModel(torch.nn.Module):
    """
        ForwardModel
        - Initial Conv1d layers for Convolution over time stacked with residual connections
        - stacked LSTM-Cells
        - Post Conv1d Layers for Convolution over time and neighbouring Mel Channels with residual connections
        - Weighting of lstm output and smoothed output
    """

    def __init__(self, input_size=30,
                 output_size=60,
                 hidden_size=180,
                 num_lstm_layers=4,
                 apply_half_sequence=True):
        super().__init__()

        self.apply_half_sequence = apply_half_sequence
        if self.apply_half_sequence:
            self.half_sequence = torch.nn.AvgPool1d(2, stride=2)
        self.lstm = torch.nn.LSTM(input_size, hidden_size, num_layers=num_lstm_layers, batch_first=True)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)

    def forward(self, x, *args):
        output, _ = self.lstm(x)
        output = self.post_linear(output)
        if self.apply_half_sequence:
            output = output.permute(0, 2, 1)
            output = self.half_sequence(output)
            output = output.permute(0, 2, 1)

        return output

########################################################################################################################
############################################### Embbedder Models  ######################################################
########################################################################################################################

class MelEmbeddingModelMelSmoothResidualUpsampling(torch.nn.Module):
    """
        EmbedderModel
        - Initial Conv1d Layers for Convolution over time and neighbouring Mel Channels with residual connections
        - stacked LSTM-Cells
        - Post upsammpling layer
    """

    def __init__(self, input_size=60,
                 output_size=300,
                 hidden_size=180,
                 num_lstm_layers=4,
                 mel_smooth_layers=3,
                 mel_smooth_filter_size=3,
                 mel_resid_activation = torch.nn.Identity(),
                 post_activation = torch.nn.LeakyReLU(),
                 post_upsampling_size=8192):
        super().__init__()

        self.mel_resid_activation = mel_resid_activation
        # self.add_vel_and_acc_info = add_vel_and_acc_info
        self.MelBlocks = torch.nn.ModuleList(
            [MelChannelConv1D(input_size, mel_smooth_filter_size) for _ in range(mel_smooth_layers)])
        self.lstm = torch.nn.LSTM(input_size, hidden_size, num_layers=num_lstm_layers, batch_first=True)
        self.post_linear = torch.nn.Linear(hidden_size, post_upsampling_size)
        self.upsampling = torch.nn.Linear(post_upsampling_size, output_size)
        self.post_activation = post_activation
        # self.resid_weighting = torch.nn.Conv1d(2 * output_size, output_size, time_filter_size, padding=2, groups=output_size)

    def forward(self, x, lens, *args):
        # IntermediateOutputs = []
        if len(self.MelBlocks) >0:
            x = x.permute(0, 2, 1)
            for layer in self.MelBlocks:
                shortcut = x
                x = layer(x)
                x += shortcut
                x = self.mel_resid_activation(x)

            x = x.permute(0, 2, 1)
        # x = self.add_vel_and_acc_info(x)
        output, (h_n, _) = self.lstm(x)
        output = torch.stack([output[i, (last - 1).long(), :] for i, last in enumerate(lens)])
        output = self.post_linear(output)
        output = self.post_activation(output)
        output = self.upsampling(output)

        return output



class EmbeddingModel(torch.nn.Module):
    """
        Embedder
        - Initial Conv1d Layers for Convolution over time and neighbouring Mel Channels with residual connections
        - stacked LSTM-Cells
        - Post upsammpling layer
    """

    def __init__(self, input_size=60,
                 output_size=300,
                 hidden_size=720,
                 num_lstm_layers=1,
                 post_activation = torch.nn.LeakyReLU(),
                 post_upsampling_size=0,
                 dropout=0):
        super().__init__()
        
        self.post_upsampling_size = post_upsampling_size
        self.lstm = torch.nn.LSTM(input_size, hidden_size, num_layers=num_lstm_layers, batch_first=True, dropout = dropout)
        if post_upsampling_size >0:
            self.post_linear = torch.nn.Linear(hidden_size, post_upsampling_size)
            self.linear_mapping = torch.nn.Linear(post_upsampling_size, output_size)
            self.post_activation = post_activation
        else:
            self.linear_mapping = torch.nn.Linear(hidden_size, output_size)
    

    def forward(self, x, lens, *args):
        output, (h_n, _) = self.lstm(x)
        output = torch.stack([output[i, (last - 1).long(), :] for i, last in enumerate(lens)])
        if self.post_upsampling_size>0:
            output = self.post_linear(output)
            output = self.post_activation(output)
        output = self.linear_mapping(output)

        return output


########################################################################################################################
################################################# Baseline Models  #####################################################
########################################################################################################################

class LinearModel(torch.nn.Module):
    def __init__(self,
                 input_channel = 30,
                 output_channel = 60,
                 mode = "inv",
                 on_full_sequence = False,
                 add_vel_and_acc=True):
        super().__init__()
        self.on_full_sequence = on_full_sequence
        self.add_vel_and_acc = add_vel_and_acc
        assert mode in ["pred", "inv", "embed"], "if you want to train a predictive model please set mode to 'pred', for a inverse model set mode to 'inv'!"
        self.mode = mode
        if self.on_full_sequence:
            if self.add_vel_and_acc:
                self.input_channel = 3 * input_channel
                self.add_vel_and_acc_info = add_vel_and_acc_info
            else:
                self.input_channel = input_channel
            if self.mode == "pred":
                self.half_sequence = torch.nn.AvgPool1d(2, stride=2)
            elif self.mode == "inv":
                self.double_sequence = double_sequence

        else:
            self.input_channel = 2*input_channel
        self.output_channel = output_channel
        self.linear = torch.nn.Linear(self.input_channel, self.output_channel)


    def forward(self, x, *args):
        if self.on_full_sequence:
            if self.add_vel_and_acc:
                x = self.add_vel_and_acc_info(x)
        else:
            x = x.reshape((x.shape[0],1,-1))

        output = self.linear(x)
        if self.on_full_sequence:
            if self.mode == "pred":
                output = output.permute(0, 2, 1)
                output = self.half_sequence(output)
                output = output.permute(0, 2, 1)
            elif self.mode == "inv":
                output = self.double_sequence(output)
        return output


class NonLinearModel(torch.nn.Module):
    def __init__(self,
                 input_channel=30,
                 output_channel=60,
                 hidden_units=8192,
                 activation_function=torch.nn.LeakyReLU(),
                 mode = "pred",
                 on_full_sequence=False,
                 add_vel_and_acc=True):
        super().__init__()
        self.on_full_sequence = on_full_sequence
        self.add_vel_and_acc = add_vel_and_acc
        assert mode in ["pred", "inv", "embed"], "if you want to train a predictive model please set mode to 'pred', for a inverse model set mode to 'inv'!"
        self.mode = mode
        if self.on_full_sequence:
            if self.add_vel_and_acc:
                self.input_channel = input_channel * 3
                self.add_vel_and_acc_info = add_vel_and_acc_info
            else:
                self.input_channel = input_channel
            if self.mode == "pred":
                self.half_sequence = torch.nn.AvgPool1d(2, stride=2)
            elif self.mode == "inv":
                self.double_sequence = double_sequence
        else:
            self.input_channel = input_channel * 2

        self.output_channel = output_channel
        self.hidden_units = hidden_units
        self.activation_function = activation_function
        self.non_linear = torch.nn.Linear(self.input_channel, self.hidden_units)
        self.linear = torch.nn.Linear(self.hidden_units, self.output_channel)

    def forward(self, x, *args):
        if self.on_full_sequence:
            if self.add_vel_and_acc:
                x = self.add_vel_and_acc_info(x)
            if self.mode == "embed":
                x = torch.sum(x, axis=1)
        else:
            x = x.reshape((x.shape[0],1, -1))
        output = self.non_linear(x)
        output = self.activation_function(output)
        output = self.linear(output)
        if self.on_full_sequence:
            if self.mode == "pred":
                output = output.permute(0, 2, 1)
                output = self.half_sequence(output)
                output = output.permute(0, 2, 1)
            elif self.mode == "inv":
                output = self.double_sequence(output)
        return output

########################################################################################################################
############################################### Generative Models  #####################################################
########################################################################################################################

class Critic(torch.nn.Module):
    def __init__(self, input_size=30,
                 embed_size=300,
                 hidden_size=180,
                 num_res_blocks=5):
        super().__init__()

        self.inital_linear = torch.nn.Linear(input_size + embed_size, hidden_size)
        self.res_blocks = torch.nn.ModuleList(
            [self._block(hidden_size, hidden_size, 5, 1, 2) for _ in range(num_res_blocks)])

    def _block(self, in_channels, out_channels, kernel_size, stride, padding):
        return torch.nn.Sequential(
            torch.nn.Conv1d(
                in_channels, out_channels, kernel_size, stride, padding,
            ),
            torch.nn.InstanceNorm1d(out_channels, affine=True),
            torch.nn.LeakyReLU(0.2),
        )

    def forward(self, x, length, vector):
        x = torch.cat([x, vector.unsqueeze(1).repeat(1, x.shape[1], 1)], dim=2)
        output = self.inital_linear(x)
        output = output.permute(0, 2, 1)

        for i, block in enumerate(self.res_blocks):
            resid = output
            output = block(output)
            output += resid

            # average pooling
        output = output.mean([1, 2])
        return output


class Generator(torch.nn.Module):
    def __init__(self, channel_noise=100,
                 embed_size=300,
                 fc_size=1024,
                 inital_seq_length=4,
                 hidden_size=256,
                 num_res_blocks=5,
                 output_size=30):
        super().__init__()

        self.fc_size = fc_size
        self.hidden_size = hidden_size
        self.fc_reshaped_size = int(fc_size / inital_seq_length)
        self.fully_connected = torch.nn.Linear(channel_noise + embed_size, fc_size)

        self.res_blocks = torch.nn.ModuleList([self._block(self.fc_reshaped_size, hidden_size, 5, 1, 2)])
        self.res_blocks = self.res_blocks.extend(
            torch.nn.ModuleList([self._block(hidden_size, hidden_size, 5, 1, 2) for _ in range(num_res_blocks - 1)]))

        self.post_linear = torch.nn.Linear(hidden_size, output_size)
        self.final_smoothing = torch.nn.Conv1d(output_size, output_size, kernel_size=5, padding=2, groups=output_size)
        self.output_activation = torch.nn.Tanh()

    def _block(self, in_channels, out_channels, kernel_size, stride, padding):
        return torch.nn.Sequential(
            torch.nn.Conv1d(
                in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding,
            ),
            torch.nn.BatchNorm1d(out_channels),
            torch.nn.LeakyReLU(0.2),
        )

    def forward(self, x, length, vector):
        x = torch.cat([x, vector.unsqueeze(1)], dim=2)
        output = self.fully_connected(x)
        output = output.view((len(x), self.fc_reshaped_size, int(output.shape[-1] / self.fc_reshaped_size)))

        for i, block in enumerate(self.res_blocks):
            size_ = int(length / (len(self.res_blocks) - i))
            resizing = torch.nn.Upsample(size=(size_), mode='linear', align_corners=False)
            output = resizing(output)
            resid = output
            output = block(output)
            if i == 0:
                if self.fc_reshaped_size == self.hidden_size:
                    output += resid
            else:
                output += resid

        output = output.permute(0, 2, 1)
        output = self.post_linear(output)
        output = output.permute(0, 2, 1)
        resid = output
        output = self.final_smoothing(output)
        output += resid
        output = output.permute(0, 2, 1)
        output = self.output_activation(output)

        return output


class SemVecToCpModel(torch.nn.Module):
    def __init__(self,
                 input_size=300, #semantic vector dim
                 output_size=30,
                 hidden_size=180,
                 num_lstm_layers=4,
                 resid_blocks=5,
                 time_filter_size=5,
                 pre_resid_activation = torch.nn.Identity(),
                 post_resid_activation = torch.nn.Identity(),
                 output_activation = torch.nn.Identity(),
                 lstm_resid=True):
        super().__init__()

        self.lstm_resid = lstm_resid
        self.pre_activation = pre_resid_activation
        self.post_activation = post_resid_activation
        self.output_activation = output_activation

        self.lstm = torch.nn.LSTM(input_size, hidden_size, num_layers=num_lstm_layers, batch_first=True)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)
        self.ResidualConvBlocks = torch.nn.ModuleList(
            [TimeConvResBlock(output_size, time_filter_size, self.pre_activation, self.post_activation) for _ in
             range(resid_blocks)])
        if self.lstm_resid and len(self.ResidualConvBlocks) > 0:
            self.resid_weighting = torch.nn.Conv1d(2 * output_size, output_size, time_filter_size, padding=2,
                                               groups=output_size)

    def forward(self,x, *args):
        output, _ = self.lstm(x)
        output = self.post_linear(output)

        output = output.permute(0, 2, 1)
        lstm_output = output
        for layer in self.ResidualConvBlocks:
            output = layer(output)

        if len(self.ResidualConvBlocks) > 0 and self.lstm_resid:
            output = [torch.stack((output[:, i, :], lstm_output[:, i, :]), axis=1) for i in range(output.shape[1])]
            output = torch.cat(output, axis=1)
            output = self.resid_weighting(output)

        output = self.output_activation(output.permute(0, 2, 1))
        return output



class SemVecToMelModel(torch.nn.Module):
    def __init__(self,
                 input_size=300,  # semantic vector dim
                 output_size=60,
                 hidden_size=180,
                 num_lstm_layers=4,
                 mel_smooth_layers=3,
                 mel_smooth_filter_size=3,
                 mel_resid_activation=torch.nn.Identity(),
                 time_filter_size=5,
                 output_activation=torch.nn.Identity(),
                 lstm_resid=True):
        super().__init__()

        self.lstm_resid = lstm_resid
        self.output_activation = output_activation
        self.mel_resid_activation = mel_resid_activation

        self.lstm = torch.nn.LSTM(input_size, hidden_size, num_layers=num_lstm_layers, batch_first=True)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)
        self.MelBlocks = torch.nn.ModuleList(
            [MelChannelConv1D(output_size, mel_smooth_filter_size) for _ in range(mel_smooth_layers)])
        if self.lstm_resid and len(self.MelBlocks) > 0:
            self.resid_weighting = torch.nn.Conv1d(2 * output_size, output_size, time_filter_size, padding=2,
                                                   groups=output_size)

    def forward(self, x, *args):
        output, _ = self.lstm(x)
        output = self.post_linear(output)

        output = output.permute(0, 2, 1)
        lstm_output = output

        for layer in self.MelBlocks:
            shortcut = output
            output = layer(output)
            output += shortcut
            output = self.mel_resid_activation(output)

        if len(self.MelBlocks) > 0 and self.lstm_resid:
            output = [torch.stack((lstm_output[:, i, :], output[:, i, :]), axis=1) for i in range(output.shape[1])]
            output = torch.cat(output, axis=1)
            output = self.resid_weighting(output)

        output = self.output_activation(output.permute(0, 2, 1))
        return output


class LSTMCritic(torch.nn.Module):
    def __init__(self, input_size=30,
                 embed_size = 300, 
                 output_size=1,
                 hidden_size=200,
                 num_lstm_layers=2,
                 dropout=0.5):
        super().__init__()
        self.lstm = torch.nn.LSTM(input_size + embed_size, hidden_size, num_layers=num_lstm_layers, batch_first = True, dropout=dropout)
        self.fully_connected = torch.nn.Linear(hidden_size,output_size)

    def forward(self, x, lens, vector, *args):
        
        x = torch.cat([x, vector.unsqueeze(1).repeat(1, x.shape[1], 1)],dim=2)
        output, (h_n, _) = self.lstm(x)
        output = torch.stack([output[i, (last - 1).long(), :] for i, last in enumerate(lens)])
        output = self.fully_connected(output)
        # average pooling
        #output = output.mean([1])
        
        return output


class LSTMGenerator(torch.nn.Module):
    def __init__(self,channel_noise = 60,
                 embed_size = 300,
                 output_size=30,
                 hidden_size=200,
                 num_lstm_layers=2,
                 dropout=0.5,
                 activation = torch.nn.LeakyReLU(0.2)):
        super().__init__()
        
        self.output_activation = torch.nn.Tanh()
        self.activation = activation
        self.fully_connected = torch.nn.Linear(channel_noise + embed_size, hidden_size)
        self.lstm = torch.nn.LSTM(hidden_size, hidden_size, num_layers=num_lstm_layers, batch_first = True, dropout=dropout)
        self.post_linear = torch.nn.Linear(hidden_size, output_size)


    def forward(self, x, lens, vector, *args):
        x = torch.cat([x,vector.unsqueeze(1).repeat(1,x.shape[1],1)], dim = 2)
        output = self.fully_connected(x)
        output = self.activation(output)
        output, _ = self.lstm(output)
        #output = torch.stack([output[i, (last - 1).long(), :] for i, last in enumerate(lens)])
        
        output = self.post_linear(output)
        output = self.output_activation(output)

        return output



import ctypes
import io
import os
import shutil
import sys
import tempfile
import zipfile

import numpy as np
import matplotlib.pyplot as plt
import librosa
import torch.nn
import pandas as pd
import requests

DIR = os.path.dirname(__file__)
_PREFIX = ''
_FILE_ENDING = ''
if sys.platform.startswith('linux'):
    _FILE_ENDING = '.so'
    _PREFIX = 'lib'
elif sys.platform.startswith('win32'):
    _FILE_ENDING = '.dll'
elif sys.platform.startswith('darwin'):
    _FILE_ENDING = '.dylib'
    _PREFIX = 'lib'

# initialize vtl
VTL = ctypes.cdll.LoadLibrary(os.path.join(DIR, f'vocaltractlab_api/{_PREFIX}VocalTractLabApi{_FILE_ENDING}'))
SPEAKER_FILE_NAME = ctypes.c_char_p(os.path.join(DIR, 'vocaltractlab_api/JD3.speaker').encode())
FAILURE = VTL.vtlInitialize(SPEAKER_FILE_NAME)
if FAILURE != 0:
    raise ValueError('Error in vtlInitialize! Errorcode: %i' % FAILURE)
del SPEAKER_FILE_NAME, FAILURE, _FILE_ENDING, _PREFIX

# get version / compile date
VERSION = ctypes.c_char_p(b' ' * 64)
VTL.vtlGetVersion(VERSION)
print('Version of the VocalTractLab library: "%s"' % VERSION.value.decode())
del VERSION


# This should be done on all cp_deltas
#np.max(np.stack((np.abs(np.min(delta, axis=0)), np.max(delta, axis=0))), axis=0)
#np.max(np.stack((np.abs(np.min(cp_param, axis=0)), np.max(cp_param, axis=0))), axis=0)

# absolute value from max / min

#Vocal tract parameters: "HX HY JX JA LP LD VS VO TCX TCY TTX TTY TBX TBY TRX TRY TS1 TS2 TS3"
#Glottis parameters: "f0 pressure x_bottom x_top chink_area lag rel_amp double_pulsing pulse_skewness flutter aspiration_strength "


cp_means = np.array([ 5.3000e-01, -5.0800e+00, -3.0000e-02, -3.7300e+00,  7.0000e-02,
                      7.3000e-01,  4.8000e-01, -5.0000e-02,  9.6000e-01, -1.5800e+00,
                      4.4600e+00, -9.3000e-01,  2.9900e+00, -5.0000e-02, -1.4600e+00,
                     -2.2900e+00,  2.3000e-01,  1.2000e-01,  1.2000e-01,  1.0720e+02,
                      4.1929e+03,  3.0000e-02,  3.0000e-02,  6.0000e-02,  1.2200e+00,
                      8.4000e-01,  5.0000e-02,  0.0000e+00,  2.5000e+01, -1.0000e+01])
cp_stds = np.array([1.70000e-01, 4.00000e-01, 4.00000e-02, 6.30000e-01, 1.20000e-01,
                    2.20000e-01, 2.20000e-01, 9.00000e-02, 4.90000e-01, 3.10000e-01,
                    3.80000e-01, 3.70000e-01, 3.50000e-01, 3.50000e-01, 4.60000e-01,
                    3.80000e-01, 6.00000e-02, 1.00000e-01, 1.80000e-01, 9.86000e+00,
                    3.29025e+03, 2.00000e-02, 2.00000e-02, 1.00000e-02, 0.00100e+00,
                    2.00000e-01, 0.00100e+00, 0.00100e+00, 0.00100e+00, 0.00100e+00])

cp_theoretical_means = np.array([ 5.00000e-01, -4.75000e+00, -2.50000e-01, -3.50000e+00,
        0.00000e+00,  1.00000e+00,  5.00000e-01,  4.50000e-01,
        5.00000e-01, -1.00000e+00,  3.50000e+00, -2.50000e-01,
        5.00000e-01,  1.00000e+00, -1.00000e+00, -3.00000e+00,
        5.00000e-01,  5.00000e-01,  0.00000e+00,  3.20000e+02,
        1.00000e+04,  1.25000e-01,  1.25000e-01,  0.00000e+00,
        1.57075e+00,  0.00000e+00,  5.00000e-01,  0.00000e+00,
        5.00000e+01, -2.00000e+01])

cp_theoretical_stds = np.array([5.00000e-01, 1.25000e+00, 2.50000e-01, 3.50000e+00, 1.00000e+00,
       3.00000e+00, 5.00000e-01, 5.50000e-01, 3.50000e+00, 2.00000e+00,
       2.00000e+00, 2.75000e+00, 3.50000e+00, 4.00000e+00, 3.00000e+00,
       3.00000e+00, 5.00000e-01, 5.00000e-01, 1.00000e+00, 2.80000e+02,
       1.00000e+04, 1.75000e-01, 1.75000e-01, 2.50000e-01, 1.57075e+00,
       1.00000e+00, 5.00000e-01, 5.00000e-01, 5.00000e+01, 2.00000e+01])

ARTICULATOR = {0: 'vocal folds',
               1: 'tongue',
               2: 'lower incisors',
               3: 'lower lip',
               4: 'other articulator',
               5: 'num articulators',
               }

min_area = 0
max_area = 15

min_length = 0.23962031463970312
max_length = 0.6217119410833707

max_incisor = 18
min_incisor = 14

max_tongue = 1
min_tongue = -1

max_velum = 1
min_velum = 0

# tube sections area 0:6, 7 Incisor position, 8 tongue tip, 9 velum opening
tube_mins = np.concatenate([np.repeat(min_area,7), np.array([min_incisor]), np.array([min_tongue]), np.array([min_velum])])
tube_maxs = np.concatenate([np.repeat(max_area,7), np.array([max_incisor]), np.array([max_tongue]), np.array([max_velum])])

tube_theoretical_means = np.mean(np.stack([tube_mins,tube_maxs]),axis=0)
tube_theoretical_stds = np.std(np.stack([tube_mins,tube_maxs]),axis=0)


def librosa_melspec(wav, sample_rate):
    wav = librosa.resample(wav, orig_sr=sample_rate, target_sr=44100,
            res_type='kaiser_best', fix=True, scale=False)
    melspec = librosa.feature.melspectrogram(y=wav, n_fft=1024, hop_length=220, n_mels=60, sr=44100, power=1.0, fmin=10, fmax=12000)
    melspec_db = librosa.amplitude_to_db(melspec, ref=0.15)
    return np.array(melspec_db.T, order='C', dtype=np.float64)


def normalize_cp(cp):
    return (cp - cp_theoretical_means) / cp_theoretical_stds


def inv_normalize_cp(norm_cp):
    return cp_theoretical_stds * norm_cp + cp_theoretical_means

def normalize_tube(tube):
    return (tube - tube_theoretical_means)/tube_theoretical_stds

def inv_normalize_tube(norm_tube):
    return norm_tube * tube_theoretical_stds + tube_theoretical_means

# -83.52182518111363
mel_mean_librosa = librosa_melspec(np.zeros(5000), 44100)[0, 0]
mel_std_librosa = abs(mel_mean_librosa)


def normalize_mel_librosa(mel):
    return (mel - mel_mean_librosa) / mel_std_librosa


def inv_normalize_mel_librosa(norm_mel):
    return mel_std_librosa * norm_mel + mel_mean_librosa


def read_cp(filename):
    with open(filename, 'rt') as cp_file:
        # skip first 6 lines
        for _ in range(6):
            cp_file.readline()
        glottis_model = cp_file.readline().strip()
        if glottis_model != 'Geometric glottis':
            print(glottis_model)
            raise ValueError(f'glottis model is not "Geometric glottis" in file {filename}')
        n_states = int(cp_file.readline().strip())
        cp_param = np.zeros((n_states, 19 + 11))
        for ii, line in enumerate(cp_file):
            kk = ii // 2
            if kk >= n_states:
                raise ValueError(f'more states saved in file {filename} than claimed in the beginning')
            # even numbers are glottis params
            elif ii % 2 == 0:
                glottis_param = line.strip()
                cp_param[kk, 19:30] = np.mat(glottis_param)
            # odd numbers are tract params
            elif ii % 2 == 1:
                tract_param = line.strip()
                cp_param[kk, 0:19] = np.mat(tract_param)
    return cp_param


def speak(cp_param):
    """
    Calls the vocal tract lab to synthesize an audio signal from the cp_param.

    Parameters
    ==========
    cp_param : np.array
        array containing the vocal and glottis parameters for each time step
        which is 110 / 44100 seoconds (roughly 2.5 ms) (seq_length, 30)

    Returns
    =======
    (signal, sampling rate) : np.array, int
        returns the signal which is number of time steps in the cp_param array
        minus one times the time step length, i. e. ``(cp_param.shape[0] - 1) *
        110 / 44100``

    """
    # get some constants
    audio_sampling_rate = ctypes.c_int(0)
    number_tube_sections = ctypes.c_int(0)
    number_vocal_tract_parameters = ctypes.c_int(0)
    number_glottis_parameters = ctypes.c_int(0)
    number_audio_samples_per_tract_state = ctypes.c_int(0)
    internal_sampling_rate = ctypes.c_double(0)

    VTL.vtlGetConstants(ctypes.byref(audio_sampling_rate),
                        ctypes.byref(number_tube_sections),
                        ctypes.byref(number_vocal_tract_parameters),
                        ctypes.byref(number_glottis_parameters),
                        ctypes.byref(number_audio_samples_per_tract_state),
                        ctypes.byref(internal_sampling_rate))

    assert audio_sampling_rate.value == 44100
    assert number_vocal_tract_parameters.value == 19
    assert number_glottis_parameters.value == 11

    number_frames = cp_param.shape[0]
    frame_steps = 110 #2.5 ms
    # within first parenthesis type definition, second initialisation
    # 2000 samples more in the audio signal for safety
    audio = (ctypes.c_double * int((number_frames-1) * frame_steps + 2000))()

    # init the arrays
    tract_params = (ctypes.c_double * (number_frames * number_vocal_tract_parameters.value))()
    glottis_params = (ctypes.c_double * (number_frames * number_glottis_parameters.value))()

    # fill in data
    tmp = np.ascontiguousarray(cp_param[:, 0:19])
    tmp.shape = (number_frames * 19,)
    tract_params[:] = tmp
    del tmp

    tmp = np.ascontiguousarray(cp_param[:, 19:30])
    tmp.shape = (number_frames * 11,)
    glottis_params[:] = tmp
    del tmp

    # Reset time-domain synthesis
    failure = VTL.vtlSynthesisReset()
    if failure != 0:
        raise ValueError(f'Error in vtlSynthesisReset! Errorcode: {failure}')

    # Call the synthesis function. It may calculate a few seconds.
    failure = VTL.vtlSynthBlock(
                    ctypes.byref(tract_params),  # input
                    ctypes.byref(glottis_params),  # input
                    number_frames,
                    frame_steps,
                    ctypes.byref(audio),  # output
                    0)
    if failure != 0:
        raise ValueError('Error in vtlSynthBlock! Errorcode: %i' % failure)

    return (np.array(audio[:-2000]), 44100)


def audio_padding(sig, samplerate, winlen=0.010):
    """
    Pads the signal by half a window length on each side with zeros.

    Parameters
    ==========
    sig : np.array
        the audio signal
    samplerate : int
        sampling rate
    winlen : float
        the window size in seconds

    """
    pad = int(np.ceil(samplerate * winlen)/2)
    z = np.zeros(pad)
    pad_signal = np.concatenate((z,sig,z))
    return pad_signal


def mel_to_sig(mel, mel_min=0.0):
    """
    creates audio from a normlised log mel spectrogram.

    Parameters
    ==========
    mel : np.array
        normalised log mel spectrogram (n_mel, seq_length)
    mel_min : float
        original min value (default: 0.0)

    Returns
    =======
    (sig, sampling_rate) : (np.array, int)

    """
    mel = mel + mel_min
    mel = inv_normalize_mel_librosa(mel)
    mel = np.array(mel.T, order='C')
    mel = librosa.db_to_amplitude(mel, ref=0.15)
    sig = librosa.feature.inverse.mel_to_audio(mel, sr=44100, n_fft=1024,
                                             hop_length=220, win_length=1024,
                                             power=1.0, fmin=10, fmax=12000)
    # there are always 110 data points missing compared to the speak function using VTL
    # add 55 zeros to the beginning and the end
    sig = np.concatenate((np.zeros(55), sig, np.zeros(55)))
    return (sig, 44100)

def array_to_tensor(array):
    """
    Creates a Tensor with batch dim = 1

    Parameters
    ==========
    array : np.array

    Returns
    =======
    torch_tensor : troch.tensor (1, ...)
    """
    torch_tensor = array.copy()
    torch_tensor.shape = (1,) + torch_tensor.shape
    torch_tensor = torch.from_numpy(torch_tensor).detach().clone()
    return torch_tensor

def speak_and_extract_tube_information(cp_param):
    """
    Calls the vocal tract lab to synthesize an audio signal from the cp_param and simultaneously extract tube information
    Parameters
    ==========
    cp_param : np.array
        array containing the vocal and glottis parameters for each time step
        which is 110 / 44100 seoconds (roughly 2.5 ms) (seq_length, 30)
    Returns
    =======
    (signal, sampling rate, tube_info) : np.array, int, dict
        returns - the signal which is number of time steps in the cp_param array
                minus one times the time step length, i. e. ``(cp_param.shape[0] - 1) *
                110 / 44100``
                - the sampling rate (44100)
                - a dictionary containing arrays for the
                tube_length_cm (seq_length, 40 tube segments),
                tube_area_cm2 (seq_length, 40 tube segments),
                tube_articulator (seq_length, 40 tube segments),
                incisor_pos_cm (seq_length,),
                tongue_tip_side_elevation (seq_length, ),
                velum_opening_cm2 (seq_length, ).

    """
    # get some constants
    audio_sampling_rate = ctypes.c_int(0)
    number_tube_sections = ctypes.c_int(0)
    number_vocal_tract_parameters = ctypes.c_int(0)
    number_glottis_parameters = ctypes.c_int(0)
    number_audio_samples_per_tract_state = ctypes.c_int(0)
    internal_sampling_rate = ctypes.c_double(0)

    VTL.vtlGetConstants(ctypes.byref(audio_sampling_rate),
                        ctypes.byref(number_tube_sections),
                        ctypes.byref(number_vocal_tract_parameters),
                        ctypes.byref(number_glottis_parameters),
                        ctypes.byref(number_audio_samples_per_tract_state),
                        ctypes.byref(internal_sampling_rate))

    assert audio_sampling_rate.value == 44100
    assert number_vocal_tract_parameters.value == 19
    assert number_glottis_parameters.value == 11

    number_frames = cp_param.shape[0]
    frame_steps = 110  # 2.5 ms
    # within first parenthesis type definition, second initialisation
    audio = [(ctypes.c_double * int(frame_steps))() for _ in range(number_frames - 1)]

    # init the arrays
    tract_params = [(ctypes.c_double * (number_vocal_tract_parameters.value))() for _ in range(number_frames)]
    glottis_params = [(ctypes.c_double * (number_glottis_parameters.value))() for _ in range(number_frames)]

    # fill in data
    tmp = np.ascontiguousarray(cp_param[:, 0:19])
    for i in range(number_frames):
        tract_params[i][:] = tmp[i]
    del tmp

    tmp = np.ascontiguousarray(cp_param[:, 19:30])
    for i in range(number_frames):
        glottis_params[i][:] = tmp[i]
    del tmp

    # tube sections
    tube_length_cm = [(ctypes.c_double * 40)() for _ in range(number_frames)]
    tube_area_cm2 = [(ctypes.c_double * 40)() for _ in range(number_frames)]
    tube_articulator = [(ctypes.c_int * 40)() for _ in range(number_frames)]
    incisor_pos_cm = [ctypes.c_double(0) for _ in range(number_frames)]
    tongue_tip_side_elevation = [ctypes.c_double(0) for _ in range(number_frames)]
    velum_opening_cm2 = [ctypes.c_double(0) for _ in range(number_frames)]

    # Reset time-domain synthesis
    failure = VTL.vtlSynthesisReset()
    if failure != 0:
        raise ValueError(f'Error in vtlSynthesisReset! Errorcode: {failure}')

    for i in range(number_frames):
        if i == 0:
            failure = VTL.vtlSynthesisAddTract(0, ctypes.byref(audio[0]),
                                               ctypes.byref(tract_params[i]),
                                               ctypes.byref(glottis_params[i]))
        else:
            failure = VTL.vtlSynthesisAddTract(frame_steps, ctypes.byref(audio[i-1]),
                                               ctypes.byref(tract_params[i]),

                                               ctypes.byref(glottis_params[i]))
        if failure != 0:
            raise ValueError('Error in vtlSynthesisAddTract! Errorcode: %i' % failure)

        # export
        failure = VTL.vtlTractToTube(ctypes.byref(tract_params[i]),
                                     ctypes.byref(tube_length_cm[i]),
                                     ctypes.byref(tube_area_cm2[i]),
                                     ctypes.byref(tube_articulator[i]),
                                     ctypes.byref(incisor_pos_cm[i]),
                                     ctypes.byref(tongue_tip_side_elevation[i]),
                                     ctypes.byref(velum_opening_cm2[i]))

        if failure != 0:
            raise ValueError('Error in vtlTractToTube! Errorcode: %i' % failure)

    audio = np.ascontiguousarray(audio)
    audio.shape = ((number_frames - 1) * frame_steps,)

    arti = [[ARTICULATOR[sec] for sec in list(tube_articulator_i)] for tube_articulator_i in list(tube_articulator)]
    incisor_pos_cm = [x.value for x in incisor_pos_cm]
    tongue_tip_side_elevation = [x.value for x in tongue_tip_side_elevation]
    velum_opening_cm2 = [x.value for x in velum_opening_cm2]

    tube_info = {"tube_length_cm": np.array(tube_length_cm),
                 "tube_area_cm2": np.array(tube_area_cm2),
                 "tube_articulator": np.array(arti),
                 "incisor_pos_cm": np.array(incisor_pos_cm),
                 "tongue_tip_side_elevation": np.array(tongue_tip_side_elevation),
                 "velum_opening_cm2": np.array(velum_opening_cm2)}

    return (audio, 44100, tube_info)


def plot_cp(cp, file_name):
    """
    Plots the trajectory of the 3 control parameters with 10 in each subplot
    ==========
    cp : np.array
        array containing the vocal and glottis parameters for each time step
        which is 110 / 44100 seoconds (roughly 2.5 ms) (seq_length, 30)
    filename: str
        filename for saving plot
    Returns
    =======
    """
    fig = plt.figure(figsize=(10, 10))
    ax1 = fig.add_axes([0.1, 0.65, 0.8, 0.3], ylim=(-3, 3))
    ax2 = fig.add_axes([0.1, 0.35, 0.8, 0.3], xticklabels=[], sharex=ax1, sharey=ax1)
    ax3 = fig.add_axes([0.1, 0.05, 0.8, 0.3], sharex=ax1, sharey=ax1)

    for ii in range(10):
        ax1.plot(cp[:, ii], label=f'param{ii:0d}')
    ax1.legend()
    for ii in range(10, 20):
        ax2.plot(cp[:, ii], label=f'param{ii:0d}')
    ax2.legend()
    for ii in range(20, 30):
        ax3.plot(cp[:, ii], label=f'param{ii:0d}')
    ax3.legend()
    fig.savefig(file_name, dpi=300)
    plt.close('all')


def plot_mel(mel, file_name):
    """
    Plots the log-mel spectrogram
    ==========
    mel : np.array
        normalised log mel spectrogram (n_mel, seq_length)
    filename: str
        filename for saving plot
    Returns
    =======
    """
    fig = plt.figure(figsize=(10, 6))
    plt.imshow(mel.T, aspect='equal', vmin=-5, vmax=20)
    fig.savefig(file_name, dpi=300)
    plt.close('all')


def stereo_to_mono(wave, which="both"):
    """
    Extract a channel from a stereo wave

    Parameters
    ==========
    wave: np.array
        Input wave data.
    which: {"left", "right", "both"} default = "both"
        if `mono`, `which` indicates whether the *left* or the *right* channel
        should be extracted, or whether *both* channels should be averaged.

    Returns
    =======
    wave: np.array

    """
    if which == "left":
        return wave[:, 0]
    if which == "right":
        return wave[:, 1]
    return (wave[:, 0] + wave[:, 1])/2


def pad_same_to_even_seq_length(seq):
    """
    Pad a sequence to an even sequence length by concatenating the last element once

    Parameters
    seq: np.array
    ==========

    Returns
    =======
    seq: np.array
    """

    if not seq.shape[0] % 2 == 0:
        return np.concatenate((seq, seq[-1:, :]), axis=0)
    else:
        return seq

def half_seq_by_average_pooling(seq):
    """
    Half a sequence by averaging adjecent points in time

    Parameters
    seq: np.array
    ==========

    Returns
    =======
    half_seq: np.array
    """
    if len(seq) % 2:
        seq = pad_same_to_even_seq_length(seq)
    half_seq = (seq[::2,:] + seq[1::2,:])/2
    return half_seq

def export_svgs(cps, path='svgs/', hop_length=5):
    """
    hop_length == 5 : roughly 80 frames per second
    hop_length == 16 : roughly 25 frames per second

    """
    n_tract_parameter = 19
    for ii in range(cps.shape[0] // hop_length):
        jj = ii * hop_length

        tract_params = (ctypes.c_double * 19)()
        tract_params[:] = cps[jj, :n_tract_parameter]

        file_name = os.path.join(path, f'tract{ii:05d}.svg')
        file_name = ctypes.c_char_p(file_name.encode())

        if not os.path.exists(path):
            os.mkdir(path)

        VTL.vtlExportTractSvg(tract_params, file_name)


class RMSELoss(torch.nn.Module):
    def __init__(self, eps=1e-6):
        super().__init__()
        self.mse = torch.nn.MSELoss()
        self.eps = eps

    def forward(self, yhat, y):
        loss = torch.sqrt(self.mse(yhat, y) + self.eps)
        return loss

rmse_loss = RMSELoss(eps=0)

def get_vel_acc_jerk(trajectory, *, lag=1):
    """
    Approximates the velocity, acceleration, jerk for the given trajectory for a given lag

    Parameters
    ==========
    trajectory : np.array
    lag : int

    Returns
    =======
    (velocity, acceleration, jerk) : np.array, np.array, np.array
        returns the approximated velocity, acceleration and jerk of the trajectory for a given lag
    """

    velocity = (trajectory[:, lag:, :] - trajectory[:, :-lag, :]) / lag
    acc = (velocity[:, 1:, :] - velocity[:, :-1, :]) / 1.0
    jerk = (acc[:, 1:, :] - acc[:, :-1, :]) / 1.0
    return velocity, acc, jerk


def cp_trajectory_loss(Y_hat, tgts):
    """
    Calculates an additive loss using the RMSE between predicted and target position, velocity, acc and jerk

    Parameters
    ==========
    Y_hat : 3D torch.tensor
        model predictions (batch, seq_length, features)
    tgts : 3D torch.tensor
        target tensor (batch, seq_length, features)

    Returns
    =======
    (loss, pos_loss, vel_loss, acc_loss, jerk_loss) : torch.tensor, torch.tensor, torch.tensor, torch.tensor, torch.tensor
        returns the summed total loss and all individual sub-losses
    """

    velocity, acc, jerk = get_vel_acc_jerk(tgts)
    velocity2, acc2, jerk2 = get_vel_acc_jerk(tgts, lag=2)
    velocity4, acc4, jerk4 = get_vel_acc_jerk(tgts, lag=4)

    Y_hat_velocity, Y_hat_acceleration, Y_hat_jerk = get_vel_acc_jerk(Y_hat)
    Y_hat_velocity2, Y_hat_acceleration2, Y_hat_jerk2 = get_vel_acc_jerk(Y_hat, lag=2)
    Y_hat_velocity4, Y_hat_acceleration4, Y_hat_jerk4 = get_vel_acc_jerk(Y_hat, lag=4)

    pos_loss = rmse_loss(Y_hat, tgts)
    vel_loss = rmse_loss(Y_hat_velocity, velocity) + rmse_loss(Y_hat_velocity2, velocity2) + rmse_loss(Y_hat_velocity4, velocity4)
    jerk_loss = rmse_loss(Y_hat_jerk, jerk) + rmse_loss(Y_hat_jerk2, jerk2) + rmse_loss(Y_hat_jerk4, jerk4)
    acc_loss = rmse_loss(Y_hat_acceleration, acc) + rmse_loss(Y_hat_acceleration2, acc2) + rmse_loss(Y_hat_acceleration4, acc4)

    loss = pos_loss + vel_loss + acc_loss + jerk_loss
    return loss, pos_loss, vel_loss, acc_loss, jerk_loss


def add_and_pad(xx, max_len, with_onset_dim=False):
    """
    Pad a sequence with last value to maximal length

    Parameters
    ==========
    xx : np.array
        sequence to pad (seq_length, features)
    max_len : int
        maximal length to be padded to
    with_onset_dim : bool
        add one features with 1 for the first time step and rest 0 to indicate
        sound onset

    Returns
    =======
    pad_seq : torch.Tensor
        padded sequence (max_len, features)

    """
    seq_length = xx.shape[0]
    if with_onset_dim:
        onset = np.zeros((seq_length, 1))
        onset[0, 0] = 1
        xx = np.concatenate((xx, onset), axis=1)  # shape len X (features +1)
    padding_size = max_len - seq_length
    padding_size = tuple([padding_size] + [1 for i in range(len(xx.shape) - 1)])
    xx = np.concatenate((xx, np.tile(xx[-1:], padding_size)), axis=0)
    return torch.from_numpy(xx)


def pad_batch_online(lens, data_to_pad, device="cpu", with_onset_dim=False):
    """
    pads and batches data into one single padded batch.

    Parameters
    ==========
    lens : 1D torch.tensor
        Tensor containing the length of each sample in data_to_pad of one batch
    data_to_pad : pd.Series
        series containing the data to pad

    Returns
    =======
    padded_data : torch.tensors
        Tensors containing the padded and stacked to one batch

    """
    max_len = int(max(lens))
    padded_data = torch.stack(list(data_to_pad.apply(
        lambda x: add_and_pad(x, max_len, with_onset_dim=with_onset_dim)))).to(device)

    return padded_data


def cps_to_ema_and_mesh(cps, file_prefix, *, path=""):
    """
    Calls the vocal tract lab to generate synthesized EMA trajectories.

    Parameters
    ==========
    cps : np.array
        2D array containing the vocal and glottis parameters for each time step
        which is 110 / 44100 seoconds (roughly 2.5 ms); first dimension is
        sequence and second is vocal tract lab parameters, i. e. (n_sequence,
        30)
    file_prefix : str
        the prefix of the files written
    path : str
        path where to put the output files

    Returns
    =======
    None : None
        all output is writen to files.

    """
    # get some constants
    audio_sampling_rate = ctypes.c_int(0)
    number_tube_sections = ctypes.c_int(0)
    number_vocal_tract_parameters = ctypes.c_int(0)
    number_glottis_parameters = ctypes.c_int(0)
    number_audio_samples_per_tract_state = ctypes.c_int(0)
    internal_sampling_rate = ctypes.c_double(0)

    VTL.vtlGetConstants(ctypes.byref(audio_sampling_rate),
                        ctypes.byref(number_tube_sections),
                        ctypes.byref(number_vocal_tract_parameters),
                        ctypes.byref(number_glottis_parameters),
                        ctypes.byref(number_audio_samples_per_tract_state),
                        ctypes.byref(internal_sampling_rate))

    assert audio_sampling_rate.value == 44100
    assert number_vocal_tract_parameters.value == 19
    assert number_glottis_parameters.value == 11

    number_frames = cps.shape[0]

    # init the arrays
    tract_params = (ctypes.c_double * (number_frames * number_vocal_tract_parameters.value))()
    glottis_params = (ctypes.c_double * (number_frames * number_glottis_parameters.value))()

    # fill in data
    tmp = np.ascontiguousarray(cps[:, 0:19])
    tmp.shape = (number_frames * 19,)
    tract_params[:] = tmp
    del tmp

    tmp = np.ascontiguousarray(cps[:, 19:30])
    tmp.shape = (number_frames * 11,)
    glottis_params[:] = tmp
    del tmp

    number_ema_points = 3
    surf = (ctypes.c_int * number_ema_points)()
    surf[:] = np.array([16, 16, 16])  # 16 = TONGUE

    vert = (ctypes.c_int * number_ema_points)()
    vert[:] = np.array([115, 225, 335])  # Tongue Back (TB) = 115; Tongue Middle (TM) = 225; Tongue Tip (TT) = 335

    if not os.path.exists(path):
        os.mkdir(path)

    failure = VTL.vtlTractSequenceToEmaAndMesh(
            ctypes.byref(tract_params), ctypes.byref(glottis_params),
            number_vocal_tract_parameters, number_glottis_parameters,
            number_frames, number_ema_points,
            ctypes.byref(surf), ctypes.byref(vert),
            path.encode(), file_prefix.encode())
    if failure != 0:
        raise ValueError('Error in vtlTractSequenceToEmaAndMesh! Errorcode: %i' % failure)


def cps_to_ema(cps):
    """
    Calls the vocal tract lab to generate synthesized EMA trajectories.

    Parameters
    ==========
    cps : np.array
        2D array containing the vocal and glottis parameters for each time step
        which is 110 / 44100 seoconds (roughly 2.5 ms); first dimension is
        sequence and second is vocal tract lab parameters, i. e. (n_sequence,
        30)

    Returns
    =======
    emas : pd.DataFrame
        returns the 3D ema points for different virtual EMA sensors in a
        pandas.DataFrame

    """
    with tempfile.TemporaryDirectory(prefix='python_paule_') as path:
        file_name = 'pyndl_util_ema_export'
        cps_to_ema_and_mesh(cps, file_prefix=file_name, path=path)
        emas = pd.read_table(os.path.join(path, f"{file_name}-ema.txt"), sep=' ')
    return emas


def seg_to_cps(seg_file):
    """
    Calls the vocal tract lab to read a segment file (seg_file) and returns the
    unnormalised cps.

    Parameters
    ==========
    seg_file : str
        path to the segment file

    Returns
    =======
    cps : np.array
        two dimensional numpy array of the unnormalised control parameter
        trajectories

    """
    segment_file_name = seg_file.encode()

    with tempfile.TemporaryDirectory() as tmpdirname:
        gesture_file_name = os.path.join(tmpdirname, 'vtl_ges_file.txt').encode()
        failure = VTL.vtlSegmentSequenceToGesturalScore(segment_file_name, gesture_file_name)
        if failure != 0:
            raise ValueError('Error in vtlSegmentSequenceToGesturalScore! Errorcode: %i' % failure)
        cps = ges_to_cps(gesture_file_name.decode())
    return cps


def ges_to_cps(ges_file):
    """
    Calls the vocal tract lab to read a gesture file (ges_file) and returns the
    unnormalised cps.

    Parameters
    ==========
    ges_file : str
        path to the gesture file

    Returns
    =======
    cps : np.array
        two dimensional numpy array of the unnormalised control parameter
        trajectories

    """
    gesture_file_name = ges_file.encode()

    with tempfile.TemporaryDirectory() as tmpdirname:
        tract_sequence_file_name = os.path.join(tmpdirname, 'vtl_tract_seq.txt').encode()
        failure = VTL.vtlGesturalScoreToTractSequence(gesture_file_name, tract_sequence_file_name)
        if failure != 0:
            raise ValueError('Error in vtlGesturalScoreToTractSequence! Errorcode: %i' % failure)

        cps = read_cp(tract_sequence_file_name.decode())
    return cps


def get_area_info_within_oral_cavity(tube_length, tube_area, *, cm_inside=7, calculate="min"):
    """
    Extracts the tube area information within the oral cavity for a given number of cm starting from the lips.

    Parameters
    ==========
    tube_length : np.array
        tube length in cm (seq_length, 40 tube segments)
    tube_area: np.array
        tube area in cm^2 (seq_length, 40 tube segments)
    cm_inside: int
        the number of cm to extend into the oral cavity
    calculate: str
        the information to extrect for each cm

    Returns
    =======
    section_area_per_time : np.array
        the extracted area information per time (seq_length, )
    """
    length_per_time = np.cumsum(tube_length, axis=1)
    section_area_per_time = []
    for t, l in enumerate(length_per_time):
        steps = [length_per_time[t][-1] - i * 1 for i in range(cm_inside + 1)][::-1]
        section_area = []
        for i, step in enumerate(steps[:-1]):
            indices = np.where(np.logical_and(l>=step, l<=steps[i+1]))[0]
            # add one more index, if possible as the next tube section is
            # paritally in this interval
            if indices[-1] < tube_area.shape[1] - 1:
                indices = np.concatenate((indices, indices[-1:] + 1))
            area = tube_area[t, indices]
            if calculate == "raw":
                section_area += [area]
            elif calculate == "mean":
                section_area += [np.mean(area)]
            elif calculate == "binary":
                section_area += [bool(np.sum(area <= 0.001))]
            elif calculate == "min":
                section_area += [np.min(area)]
            else:
                raise Exception(f"calculate must be one of ['raw', 'mean', 'binary', 'min']")
        section_area_per_time += [section_area]
    return np.asarray(section_area_per_time)


def download_pretrained_weights(*, skip_if_exists=True, verbose=True):
    package_path = DIR
    model_weights_path = os.path.join(package_path, 'pretrained_models')
    if os.path.isdir(model_weights_path):
        if skip_if_exists:
            if verbose:
                print(f"pretrained_models exist already. Skip download. Path is {model_weights_path}")
                print(f'Version of pretrained weights is "{get_pretrained_weights_version()}"')
            return
        shutil.rmtree(model_weights_path)

    zip_file_url = "https://nc.mlcloud.uni-tuebingen.de/index.php/s/N4nik8wgxwQHP83/download"
    if verbose:
        print(f"downloading 200 MB of pretrained weights from {zip_file_url}")
        print(f"saving pretrained weights to {model_weights_path}")
    stream = requests.get(zip_file_url, stream=True)
    zip_file = zipfile.ZipFile(io.BytesIO(stream.content))
    zip_file.extractall(package_path)
    if verbose:
        print(f'Version of pretrained weights is "{get_pretrained_weights_version()}"')


def get_pretrained_weights_version():
    """read and return the version of the pretrained weights, <No version file
    found> if no pretrained weights exist"""
    version_path = os.path.join(DIR, 'pretrained_models/version.txt')
    if not os.path.exists(version_path):
        return f"<No version file found at {version_path}>"
    with open(version_path, 'rt') as vfile:
        version = vfile.read().strip()
    return version




"""
Here we are doing gradient based planning on a target utterance (target mel spec).
For that we need an already trained predictive model (``pred_model``) that
predicts a mel spec from control parameters (cp) and an inverse model
(``inv_model``) for initialisation that generates cps from mel spec. In this
version we use a third model that maps a mel spec onto a fixed length semantic
vector (``embedder``).
The general idea works as follows:
    1. Initialize the cp for the full utterance time steps with the inverse model and
       the target mel spec. We initialize the target semvec by making a forward
       pass through the embedder.
    2. Predict the full utterance with the predictive model using the
       initialized cp and predict the semvecs from the predicted acoustics.
    3. Compare the predictions with the target mel spec to create a
       discrepancy. Do the same for the target semvec and the predicted semvec.
    4. Plan by using the discrepancies from acoustics and semvecs and jerk and
       velocity loss on the cps to update the cps.
    5. Do this 200 times
    6. after 200 planning updates continue the training of the predictive model
       with an actual synthesized version

"""

import random
import time
import os
from collections import namedtuple

import pandas as pd
import numpy as np
from tqdm import tqdm

import torch
from torch.nn import L1Loss, MSELoss
import soundfile as sf

# Set seed
torch.manual_seed(20200905)
random.seed(20200905)

tqdm.pandas()

from .util import (speak, inv_normalize_cp, normalize_mel_librosa,
        stereo_to_mono, librosa_melspec, RMSELoss, get_vel_acc_jerk,
        cp_trajectory_loss, mel_to_sig, pad_batch_online,
        speak_and_extract_tube_information, normalize_tube,
        get_area_info_within_oral_cavity, get_pretrained_weights_version)

from .models import (ForwardModel, InverseModelMelTimeSmoothResidual, EmbeddingModel, Generator, NonLinearModel)

from . import visualize

DIR = os.path.dirname(__file__)


PlanningResults = namedtuple('PlanningResults', "planned_cp, initial_cp, initial_sig, initial_sr, initial_prod_mel,initial_pred_mel, target_sig, target_sr, target_mel, prod_sig, prod_sr, prod_mel, pred_mel, initial_prod_semvec, initial_pred_semvec, prod_semvec, pred_semvec, prod_loss_steps, planned_loss_steps, planned_mel_loss_steps, vel_loss_steps, jerk_loss_steps, pred_semvec_loss_steps, prod_semvec_loss_steps, cp_steps, pred_semvec_steps, prod_semvec_steps, grad_steps, sig_steps, prod_mel_steps, pred_mel_steps, pred_model_loss, inv_model_loss")
PlanningResultsWithSomatosensory = namedtuple('PlanningResultsWithSomatosensory', "planned_cp, initial_cp, initial_sig, initial_sr, initial_prod_mel,initial_pred_mel, initial_prod_tube, initial_pred_tube, initial_prod_tube_mel, initial_pred_tube_mel, target_sig, target_sr, target_mel, prod_sig, prod_sr, prod_mel, pred_mel, prod_tube, pred_tube, prod_tube_mel, pred_tube_mel, initial_prod_semvec, initial_pred_semvec, initial_prod_tube_semvec, initial_pred_tube_semvec, prod_semvec, pred_semvec, prod_tube_semvec, pred_tube_semvec, prod_loss_steps, planned_loss_steps, planned_mel_loss_steps, vel_loss_steps, jerk_loss_steps, pred_semvec_loss_steps, prod_semvec_loss_steps, prod_tube_loss_steps, pred_tube_mel_loss_steps,prod_tube_mel_loss_steps, pred_tube_semvec_loss_steps, prod_tube_semvec_loss_steps, cp_steps, pred_semvec_steps, prod_semvec_steps, grad_steps, sig_steps, prod_mel_steps, pred_mel_steps, prod_tube_steps, pred_tube_steps, prod_tube_mel_steps, pred_tube_mel_steps, prod_tube_semvec_steps, pred_tube_semvec_steps, pred_model_loss, inv_model_loss, tube_model_loss, tube_mel_model_loss")

BestSynthesisAcoustic = namedtuple('BestSynthesisAcoustic', "mel_loss, planned_cp, prod_sig, prod_mel, pred_mel")
BestSynthesisSemantic = namedtuple('BestSynthesisSemantic', "semvec_loss, planned_cp, prod_sig, prod_semvec, pred_semvec")
BestSynthesisSomatosensory = namedtuple('BestSynthesisSomatosensory',"tube_loss, tube_mel_loss, tube_semvec_loss, planned_cp, prod_sig, prod_tube, pred_tube, prod_tube_mel, pred_tube_mel, prod_tube_semvec, pred_tube_semvec")

rmse_loss = RMSELoss(eps=0)
l2 = MSELoss()
l1 = L1Loss()


def velocity_jerk_loss(pred, loss, *, guiding_factor=None):
    """returns (velocity_loss, jerk_loss) tuple"""
    vel1, acc1, jerk1 = get_vel_acc_jerk(pred)
    vel2, acc2, jerk2 = get_vel_acc_jerk(pred, lag=2)
    vel4, acc4, jerk4 = get_vel_acc_jerk(pred, lag=4)

    loss = rmse_loss

    # in the lag calculation higher lags are already normalised to standard
    # units
    if guiding_factor is None:
        velocity_loss = (loss(vel1, torch.zeros_like(vel1))
                         + loss(vel2, torch.zeros_like(vel2))
                         + loss(vel4, torch.zeros_like(vel4)))
        jerk_loss = (loss(jerk1, torch.zeros_like(jerk1))
                     + loss(jerk2, torch.zeros_like(jerk2))
                     + loss(jerk4, torch.zeros_like(jerk4)))
    else:
        assert 0.0 < guiding_factor < 1.0
        velocity_loss = (loss(vel1, guiding_factor * vel1.detach().clone())
                         + loss(vel2, guiding_factor * vel2.detach().clone())
                         + loss(vel4, guiding_factor * vel4.detach().clone()))
        jerk_loss = (loss(jerk1, guiding_factor * jerk1.detach().clone())
                     + loss(jerk2, guiding_factor * jerk2.detach().clone())
                     + loss(jerk4, guiding_factor * jerk4.detach().clone()))

    return velocity_loss, jerk_loss



class Paule():
    """
    The class paule keeps track of the the state associated with the Predictive
    Articulatory speech synthesise Using Lexical Embeddings.
    This state especially are the weights of the predictive, inverse and
    embedder model as well as data used for continue learning.

    """

    def __init__(self, *, pred_model=None, pred_optimizer=None, inv_model=None, inv_optimizer=None,
                 embedder=None, cp_gen_model=None, mel_gen_model=None,
                 use_somatosensory_feedback=False, cp_tube_model=None, tube_optimizer=None,
                 tube_mel_model=None, tube_mel_optimizer=None, tube_embedder=None,
                 continue_data=None, device=torch.device('cpu'), smiling=False):

        # load the pred_model, inv_model and embedder here
        # for cpu
        self.device = device

        self.smiling = smiling

        print(f'Version of pretrained weights is "{get_pretrained_weights_version()}"')

        # PREDictive MODEL (cp -> mel)
        if pred_model:
            self.pred_model = pred_model
        else:
            self.pred_model = ForwardModel(num_lstm_layers=1, hidden_size=720).double()
            self.pred_model.load_state_dict(
                torch.load(os.path.join(DIR, "pretrained_models/predictive/pred_model_common_voice_1_720_lr_0001_50_00001_50_000001_50_0000001_200.pt"),
                           map_location=self.device))
            # Non-Linear Perceptron PREDictive MODEL
            #self.pred_model = NonLinearModel(input_channel=30, output_channel=60,
            #                         mode="pred",
            #                         hidden_units=30000,
            #                         on_full_sequence=True,
            #                         add_vel_and_acc=False).double()
            #self.pred_model.load_state_dict(
            #    torch.load(os.path.join(DIR,
            #                            "pretrained_models/predictive/pred_leaky_relu_non_linear_30000_hidden_lr_0001_50_00001_50_000001_50_0000001_200.pt"),
            #               map_location=self.device))

        self.pred_model = self.pred_model.to(self.device)

        # INVerse MODEL (mel -> cp)
        if inv_model:
            self.inv_model = inv_model
        else:
            self.inv_model = InverseModelMelTimeSmoothResidual(num_lstm_layers=1, hidden_size=720).double()
            self.inv_model.load_state_dict(
                torch.load(os.path.join(DIR, "pretrained_models/inverse/inv_model_common_voice_3_1_720_5_lr_0001_50_00001_50_000001_50_0000001_200.pt"),
                           map_location=self.device))
            # Non-Linear Perceptron INVerse MODEL
            #self.inv_model = NonLinearModel(input_channel=60, output_channel=30,
            #                                 mode="inv",
            #                                 hidden_units=30000,
            #                                 on_full_sequence=True,
            #                                 add_vel_and_acc=False).double()
            #self.inv_model.load_state_dict(
            #    torch.load(os.path.join(DIR,
            #                            "pretrained_models/inverse/inv_leaky_relu_non_linear_30000_hidden_lr_0001_50_00001_50_000001_50_0000001_200.pt"),
            #               map_location=self.device))
        self.inv_model = self.inv_model.to(self.device)

        # EMBEDDER (mel -> semvec)
        if embedder:
            self.embedder = embedder
        else:
            self.embedder = EmbeddingModel(num_lstm_layers=2, hidden_size=720).double()
            self.embedder.load_state_dict(torch.load(
                os.path.join(DIR, "pretrained_models/embedder/embed_model_common_voice_syn_rec_2_720_0_dropout_07_noise_6e05_rmse_lr_00001_200.pt"),
                map_location=self.device))
            self.embedder.eval()
            # Non-Linear Perceptron Embedder
            #self.embedder = NonLinearModel(input_channel=60, output_channel=300,
            #                                mode="embed",
            #                                hidden_units=30000,
            #                                on_full_sequence=True,
            #                                add_vel_and_acc=False).double()
            #self.embedder.load_state_dict(
            #    torch.load(os.path.join(DIR,
            #                            "pretrained_models/embedder/model_embed_sum_leaky_relu_non_linear_30000_hidden_noise_6e05_lr_00001_200.pt"),
            #               map_location=self.device))

        self.embedder = self.embedder.to(self.device)

        # CP GENerative MODEL
        if cp_gen_model:
            self.cp_gen_model = cp_gen_model
        else:
            self.cp_gen_model = Generator().double()
            self.cp_gen_model.load_state_dict(torch.load(
                os.path.join(DIR, "pretrained_models/cp_gan/conditional_trained_cp_generator_whole_critic_it_5_10_20_40_80_100_415.pt"),
                map_location=self.device))
        self.cp_gen_model = self.cp_gen_model.to(self.device)
        self.cp_gen_model.eval()

        # MEL GENerative MODEL
        if mel_gen_model:
            self.mel_gen_model = mel_gen_model
        else:
            self.mel_gen_model = Generator(output_size=60).double()
            self.mel_gen_model.load_state_dict(torch.load(
                os.path.join(DIR, "pretrained_models/mel_gan/conditional_trained_mel_generator_synthesized_critic_it_5_10_20_40_80_100_400.pt"),
                                                          map_location=self.device))
        self.mel_gen_model = self.mel_gen_model.to(self.device)
        self.mel_gen_model.eval()

        self.use_somatosensory_feedback = use_somatosensory_feedback
        if self.use_somatosensory_feedback:
            # CP-Tube Model (cp -> tube)
            if cp_tube_model:
                self.cp_tube_model = cp_tube_model
            else:
                self.cp_tube_model = ForwardModel(num_lstm_layers=1,
                                             hidden_size=360,
                                             output_size=10,
                                             input_size=30,
                                             apply_half_sequence=False).double()
                self.cp_tube_model.load_state_dict(torch.load(
                    os.path.join(DIR, "pretrained_models/somatosensory/cp_to_tube_model_1_360_lr_0001_50_00001_100.pt"),
                                                         map_location=self.device))
            self.cp_tube_model = self.cp_tube_model.to(self.device)

            # Tube-Mel Model (tube -> mel)
            if tube_mel_model:
                self.tube_mel_model = tube_mel_model
            else:
                self.tube_mel_model = ForwardModel(num_lstm_layers=1,
                     hidden_size = 360,
                     output_size = 60,
                    input_size = 10,
                    apply_half_sequence=True).double()
                self.tube_mel_model.load_state_dict(torch.load(
                    os.path.join(DIR, "pretrained_models/somatosensory/tube_to_mel_model_1_360_lr_0001_50_00001_100.pt"),
                                                          map_location=self.device))
            self.tube_mel_model = self.tube_mel_model.to(self.device)

            # Tube-Embedder Model (tube -> semvec)
            if tube_embedder:
                self.tube_embedder = tube_embedder
            else:
                self.tube_embedder = EmbeddingModel(input_size = 10,
                                                    num_lstm_layers=2,
                                                    hidden_size=720,
                                                    dropout=0.7,
                                                    post_upsampling_size=0).double()
                self.tube_embedder.load_state_dict(torch.load(
                    os.path.join(DIR, "pretrained_models/somatosensory/tube_to_vector_model_2_720_0_dropout_07_noise_6e05_rmse_lr_00001_200.pt"),
                           map_location=self.device))
            self.tube_embedder = self.tube_embedder.to(self.device)
            self.tube_embedder.eval()


        # DATA to continue learning
        self.continue_data = continue_data
        self.continue_data_limit = 1000  # max amount of training data stored in paule instance

        if self.continue_data is not None:
            if len(self.continue_data) > self.continue_data_limit:
                random_sample = random.sample(range(len(self.continue_data)), self.continue_data_limit)
                self.continue_data = self.continue_data.iloc[random_sample].reset_index(drop=True)

        if pred_optimizer:
            self.pred_optimizer = pred_optimizer
        else:
            self.pred_optimizer = torch.optim.Adam(self.pred_model.parameters(), lr=0.001)
        self.pred_criterion = rmse_loss

        if inv_optimizer:
            self.inv_optimizer = inv_optimizer
        else:
            self.inv_optimizer = torch.optim.Adam(self.inv_model.parameters(), lr=0.001)
        self.inv_criterion = cp_trajectory_loss

        if self.use_somatosensory_feedback:
            if tube_optimizer:
                self.tube_optimizer = tube_optimizer
            else:
                self.tube_optimizer = torch.optim.Adam(self.cp_tube_model.parameters(), lr=0.001)
            self.tube_criterion = rmse_loss
            if tube_mel_optimizer:
                self.tube_mel_optimizer = tube_optimizer
            else:
                self.tube_mel_optimizer = torch.optim.Adam(self.tube_mel_model.parameters(), lr=0.001)
            self.tube_mel_criterion = rmse_loss

        self.best_synthesis_acoustic = None
        self.best_synthesis_semantic = None
        if self.use_somatosensory_feedback:
            self.best_synthesis_somatosensory = None

    def create_epoch_batches(self, df_length, batch_size, shuffle=True, same_size_batching=False, sorted_training_length_keys=None, training_length_dict=None):
        """
        Create Epoch by batching indices of dataset

        Parameters
        ==========
        df_length : int
            total number of samples in training set
        batch_size : int
            number of samples in one batch
        shuffle : bool
            keep order of training set or random shuffle
        same_size_batching : bool
            create epoch of batches with similar long samples to avoid long padding
        sorted_training_length_keys : np.array
            sorted array of unique sequence lengths in training data
        training_length_dict: dict
            dictionary keys containing indices of samples with unique sequence lengths in training data and values
            containing indices of samples with corresponding length

        Returns
        =======
        epoch : list of list
            list of lists containing indices for each batch in one epoch
        """

        if same_size_batching and training_length_dict is None:
            raise ValueError("Dictionary containing indices of samples with corresponding length needed for same_size_batching!")

        if same_size_batching:
            epoch = []  # list of batches
            foundlings = []  # rest samples for each length which do not fit into one batch
            sorted_training_length_keys = np.sort(list(training_length_dict.keys())) # sorted unique lengths in training
            for length in sorted_training_length_keys:  # iterate over each unique length in training data
                length_idxs = training_length_dict[length]  # dictionary containing indices of samples with length
                rest = len(length_idxs) % batch_size
                random.shuffle(length_idxs)  # shuffle indices
                epoch += [length_idxs[i * batch_size:(i * batch_size) + batch_size] for i in
                          range(int(len(length_idxs) / batch_size))]  # cut into batches and append to epoch
                if rest > 0:
                    foundlings += list(length_idxs[
                                       -rest:])  # remaining indices which do not fit into one batch are stored in foundling
            foundlings = np.asarray(foundlings)
            rest = len(foundlings) % batch_size
            epoch += [foundlings[i * batch_size:(i * batch_size) + batch_size] for i in
                      range(int(len(
                          foundlings) / batch_size))]  # cut foudnlings into batches (because inserted sorted this ensures minimal padding)
            if rest > 0:
                epoch += [foundlings[-rest:]]  # put rest into one batch (allow smaller batch)
            random.shuffle(epoch)

        else:
            rest = df_length % batch_size
            idxs = list(range(df_length))
            if shuffle:
                random.shuffle(idxs)  # shuffle indicees
            if rest > 0:
                idxs += idxs[:(
                            batch_size - rest)]  # rolling batching (if number samples not divisible by batch_size append first again)
            epoch = [idxs[i * batch_size:(i * batch_size) + batch_size] for i in
                     range(int(len(idxs) / batch_size))]  # cut into batches
        return epoch

    def plan_iterative(self, *,
                       target_acoustic=None,
                       target_semvecs=None,
                       target_seq_lengths=None,
                       overlap=8, **kwargs):
        pass


    def plan_resynth(self, *, learning_rate_planning=0.01, learning_rate_learning=0.001,
                     learning_rate_learning_inv=None,
                     target_acoustic=None,
                     target_semvec=None,
                     target_seq_length=None,
                     initial_cp=None,
                     past_cp=None,
                     initialize_from="acoustic",
                     objective="acoustic",
                     n_outer=5, n_inner=24,
                     continue_learning=True,
                     continue_learning_inv=False,
                     continue_learning_tube=False,
                     add_training_data_pred=False,
                     add_training_data_inv=False,
                     n_batches=3, batch_size=8, n_epochs=10,
                     log_ii=1,
                     log_semantics=True,
                     log_gradients=False,
                     log_signals = False,
                     log_cps = False,
                     plot=False,
                     seed=None,
                     verbose=True):
        """
        plans resynthesis cp trajectories.

        Parameters
        ==========
        learning_rate_planning : float
            learning rate for updating cps
        learning_rate_learning : float
            learning rate for updating predictive model
        learning_rate_learning_inv: float
            learning rate for updating inverse model
        target_acoustic : str
            (target_sig, target_sr)
        target_semvec : torch.tensor
        target_seq_length : int (None)
        initial_cp : torch.tensor
        past_cp : torch.tensor
            cp that are already executed, but should be conditioned on they
            will be concatenated to the beginning of the initial_cp
        initialize_from : {'semvec', 'acoustic', None}
            can be None, if initial_cp are given
        objective : {'acoustic_semvec', 'acoustic', 'semvec'}
        n_outer : int (40)
        n_inner : int (100)
        continue_learning : bool (True)
            update predictive model with synthesized acoustics
        continue_learning_inv : bool (False)
            update inverse model
        continue_learning_tube : bool (False),
             update tube model with tube information after synthesizing acoustics
        add_training_data_pred : bool (False)
            update predictive model solely on produced acoustics during training or add training data
        add_training_data_inv : bool (False)
            update inverse model solely on produced acoustics during training or add training data
        n_batches : int
            number of batches to train on
        batch_size : int
            number of samples in one batch
        n_epochs : int
            number of epochs to train on with number of batches
        log_ii : int
            log results and synthesize audio after ii number of inner iterations
        plot : bool  or str (False)
            if False no plotting; if a string is given this is used as the path
            where to store the plots; if True interactive blocking plotting is enabled
        seed : int random seed
        verbose : bool (False)

        """

        if seed:
            torch.manual_seed(seed)
            random.seed(seed)

        if target_acoustic is None and target_semvec is None:
            raise ValueError("Either target_acoustic or target_semvec has to be not None.")

        if learning_rate_learning:
            for param_group in self.pred_optimizer.param_groups:
                param_group['lr'] = learning_rate_learning # set learning rate

        if learning_rate_learning_inv:
            for param_group in self.inv_optimizer.param_groups:
                param_group['lr'] = learning_rate_learning_inv # set learning rate

        if log_ii is None:
            log_ii = n_inner

        if log_ii > n_inner:
            raise ValueError('results can only be logged between first and last planning step')

        if isinstance(target_acoustic, str):
            target_sig, target_sr = sf.read(target_acoustic)
            if len(target_sig.shape) == 2:
                target_sig = stereo_to_mono(target_sig)
            # assert target_sr == 44100, 'sampling rate of wave name must be 44100'

        elif target_acoustic is None:
            pass
        else:
            target_sig, target_sr = target_acoustic

        if target_acoustic is None and (target_seq_length is None or target_semvec is None):
            raise ValueError("if target_acoustic is None you need to give a target_seq_length and a target_semvec")

        elif target_acoustic is None:
            mel_gen_noise = torch.randn(1, 1, 100).to(self.device)
            if not isinstance(target_semvec, torch.Tensor):
                target_semvec = torch.tensor(target_semvec)
            mel_gen_semvec = target_semvec.view(1, 300).detach().clone()
            target_mel = self.mel_gen_model(mel_gen_noise, target_seq_length, mel_gen_semvec)
            target_mel = target_mel.detach().clone()
            target_sig, target_sr = mel_to_sig(target_mel.view(target_mel.shape[1], target_mel.shape[2]).cpu().numpy())
        else:
            target_mel = librosa_melspec(target_sig, target_sr)
            target_mel = normalize_mel_librosa(target_mel)
            target_mel -= target_mel.min()
            target_mel.shape = (1,) + target_mel.shape
            target_mel = torch.from_numpy(target_mel)
            target_seq_length = target_mel.shape[1]

        target_mel = target_mel.to(self.device)

        if target_semvec is None:
            with torch.no_grad():
                target_semvec = self.embedder(target_mel, (torch.tensor(target_mel.shape[1]),))
        else:
            if not isinstance(target_semvec, torch.Tensor):
                target_semvec = torch.tensor(target_semvec)
            target_semvec = target_semvec.view(1, 300).detach().clone()
        target_semvec = target_semvec.to(self.device)

        # def constrained_criterion(xx_new, target_mel):
        #    pred_mel = self.pred_model(xx_new)
        #    mel_loss = rmse_loss(pred_mel,target_mel)
        #    velocity_loss, jerk_loss = velocity_jerk_loss(xx_new)
        #    loss = mel_loss + velocity_loss + jerk_loss
        #    return loss, mel_loss, velocity_loss, jerk_loss, pred_mel

        # 1.1 predict initial_cp
        if initial_cp is None:
            if initialize_from == "acoustic":
                xx = target_mel.detach().clone().to(self.device)
                with torch.no_grad():
                    initial_cp = self.inv_model(xx)
                initial_cp = initial_cp.detach().cpu().numpy().clip(min=-1, max=1)
                initial_cp.shape = (initial_cp.shape[1], initial_cp.shape[2])
                del xx
            elif initialize_from == "semvec":
                cp_gen_noise = torch.randn(1, 1, 100).to(self.device)
                if not isinstance(target_semvec, torch.Tensor):
                    target_semvec = torch.tensor(target_semvec)
                cp_gen_semvec = target_semvec.view(1, 300).detach().clone()
                initial_cp = self.cp_gen_model(cp_gen_noise, 2 * target_seq_length, cp_gen_semvec)
                initial_cp = initial_cp.detach().cpu().numpy()
                initial_cp.shape = (initial_cp.shape[1], initial_cp.shape[2])
            else:
                raise ValueError("initialize_from has to be either 'acoustic' or 'semvec'")

        else:
            if initialize_from is not None:
                raise ValueError('one of initial_cp and initialize_from has to be None')
            if not initial_cp.shape[0] == (target_mel.shape[1] * 2):
                raise ValueError(f"initial_cp {initial_cp.shape[0]}, target_mel {target_mel.shape[1] * 2}")

        if not past_cp is None and past_cp.shape[0] % 2 != 0:
            raise ValueError("past_cp have to be None or the sequence length has to be an even number")
        # 1.3 create initial xx
        # initial_cp = np.zeros_like(initial_cp)
        if not past_cp is None:
            initial_cp = np.concatenate((past_cp, initial_cp), axis=0)
            # store past_cp as torch tensor to reset after each planning iteration
            past_cp_torch = torch.from_numpy(past_cp)
            past_cp_torch = past_cp_torch.to(self.device)

        xx_new = initial_cp.copy()
        xx_new.shape = (1, xx_new.shape[0], xx_new.shape[1])
        xx_new = torch.from_numpy(xx_new)
        xx_new = xx_new.to(self.device)
        xx_new.requires_grad_()
        xx_new.retain_grad()


        if objective == 'acoustic_semvec':
            if self.use_somatosensory_feedback:
                def criterion(pred_mel, target_mel, pred_semvec, target_semvec, cps, pred_tube_mel, pred_tube_semvec):
                    mel_loss = rmse_loss(pred_mel, target_mel)
                    semvec_loss = rmse_loss(pred_semvec, target_semvec)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    tube_mel_loss = rmse_loss(pred_tube_mel, target_mel)
                    tube_semvec_loss = rmse_loss(pred_tube_semvec, target_semvec)
                    velocity_loss = 4 * velocity_loss
                    jerk_loss = 4 * jerk_loss
                    semvec_loss = 10 * semvec_loss
                    tube_mel_loss = tube_mel_loss
                    tube_semvec_loss = 10 * tube_semvec_loss
                    loss = mel_loss + velocity_loss + jerk_loss + semvec_loss + tube_mel_loss + tube_semvec_loss
                    return loss, mel_loss, velocity_loss, jerk_loss, semvec_loss, tube_mel_loss, tube_semvec_loss
            else:
                def criterion(pred_mel, target_mel, pred_semvec, target_semvec, cps):
                    mel_loss = rmse_loss(pred_mel, target_mel)
                    semvec_loss = rmse_loss(pred_semvec, target_semvec)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    velocity_loss = 2 * velocity_loss
                    jerk_loss = 2 * jerk_loss
                    semvec_loss = 10 * semvec_loss
                    loss = mel_loss + velocity_loss + jerk_loss + semvec_loss
                    return loss, mel_loss, velocity_loss, jerk_loss, semvec_loss

        elif objective == 'acoustic':
            if self.use_somatosensory_feedback:
                def criterion(pred_mel, target_mel, cps, pred_tube_mel):
                    mel_loss = rmse_loss(pred_mel, target_mel)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    tube_mel_loss = rmse_loss(pred_tube_mel, target_mel)
                    velocity_loss = 4 * velocity_loss
                    jerk_loss = 4 * jerk_loss
                    tube_mel_loss = tube_mel_loss
                    loss = mel_loss + velocity_loss + jerk_loss + tube_mel_loss
                    return loss, mel_loss, velocity_loss, jerk_loss, tube_mel_loss
            else:
                def criterion(pred_mel, target_mel, cps):
                    mel_loss = rmse_loss(pred_mel, target_mel)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    velocity_loss = 2 * velocity_loss
                    jerk_loss = 2 * jerk_loss
                    loss = mel_loss + velocity_loss + jerk_loss
                    return loss, mel_loss, velocity_loss, jerk_loss

        elif objective == 'semvec':
            if self.use_somatosensory_feedback:
                def criterion(pred_semvec, target_semvec, cps, pred_tube_semvec):
                    semvec_loss = rmse_loss(pred_semvec, target_semvec)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    tube_semvec_loss = rmse_loss(pred_tube_semvec, target_semvec)
                    velocity_loss = 4 * velocity_loss
                    jerk_loss = 4 * jerk_loss
                    semvec_loss = 10 * semvec_loss
                    tube_semvec_loss = 10 * tube_semvec_loss
                    loss = velocity_loss + jerk_loss + semvec_loss + tube_semvec_loss
                    return loss, velocity_loss, jerk_loss, semvec_loss, tube_semvec_loss
            else:
                def criterion(pred_semvec, target_semvec, cps):
                    semvec_loss = rmse_loss(pred_semvec, target_semvec)
                    velocity_loss, jerk_loss = velocity_jerk_loss(cps, rmse_loss)
                    velocity_loss = 2 * velocity_loss
                    jerk_loss = 2 * jerk_loss
                    semvec_loss = 10 * semvec_loss
                    loss = velocity_loss + jerk_loss + semvec_loss
                    return loss, velocity_loss, jerk_loss, semvec_loss

        else:
            raise ValueError("objective has to be one of 'acoustic_semvec', 'acoustic' or 'semvec'")

        # 1.0 create variables for logging
        prod_loss_steps = list()
        planned_loss_steps = list()
        planned_mel_loss_steps = list()
        vel_loss_steps = list()
        jerk_loss_steps = list()
        pred_semvec_loss_steps = list()
        prod_semvec_loss_steps = list()

        cp_steps = list()
        pred_semvec_steps = list()
        prod_semvec_steps = list()
        grad_steps = list()
        sig_steps = list()
        pred_mel_steps = list()
        prod_mel_steps = list()
        pred_model_loss = list()
        inv_model_loss = list()

        optimizer = torch.optim.Adam([xx_new], lr=learning_rate_planning)

        if self.use_somatosensory_feedback:
            prod_tube_loss_steps = list()
            pred_tube_mel_loss_steps = list()
            prod_tube_mel_loss_steps = list()  # predicted from tube info after calling speak
            pred_tube_semvec_loss_steps = list()
            prod_tube_semvec_loss_steps = list()  # predicted from tube info after calling speak

            pred_tube_steps = list()
            prod_tube_steps = list()
            prod_tube_mel_steps = list()
            pred_tube_mel_steps = list()
            pred_tube_semvec_steps = list()
            prod_tube_semvec_steps = list()

            tube_model_loss = list()
            tube_mel_model_loss = list()


        # initial results
        with torch.no_grad():
            initial_pred_mel = self.pred_model(xx_new)
            initial_pred_semvec = self.embedder(initial_pred_mel, (torch.tensor(initial_pred_mel.shape[1]),))

        xx_new_numpy = xx_new[-1, :, :].detach().cpu().numpy().copy()

        if self.use_somatosensory_feedback:
            with torch.no_grad():
                initial_pred_tube = self.cp_tube_model(xx_new)
                initial_pred_tube_mel = self.tube_mel_model(initial_pred_tube)
                initial_pred_tube_semvec = self.tube_embedder(initial_pred_tube,(torch.tensor(initial_pred_tube.shape[1]),))

            initial_sig, initial_sr, initial_tube_info = speak_and_extract_tube_information(inv_normalize_cp(xx_new_numpy))

            area_within_oral_cavity = get_area_info_within_oral_cavity(initial_tube_info["tube_length_cm"], initial_tube_info["tube_area_cm2"])
            initial_prod_tube = np.concatenate([area_within_oral_cavity,
                                                np.expand_dims(initial_tube_info["incisor_pos_cm"],axis=1),
                                                np.expand_dims(initial_tube_info["tongue_tip_side_elevation"],axis=1),
                                                np.expand_dims(initial_tube_info["velum_opening_cm2"],axis=1)],axis=1)
            initial_prod_tube = normalize_tube(initial_prod_tube)

            initial_prod_tube.shape = initial_pred_tube.shape
            initial_prod_tube = torch.from_numpy(initial_prod_tube)
            initial_prod_tube = initial_prod_tube.to(self.device)

            with torch.no_grad():
                initial_prod_tube_mel = self.tube_mel_model(initial_prod_tube)
                initial_prod_tube_semvec = self.tube_embedder(initial_prod_tube, (torch.tensor(initial_prod_tube.shape[1]),))

            initial_prod_tube = initial_prod_tube[-1, :, :].detach().cpu().numpy().copy()
            initial_pred_tube = initial_pred_tube[-1, :, :].detach().cpu().numpy().copy()
            initial_prod_tube_mel = initial_prod_tube_mel[-1, :, :].detach().cpu().numpy().copy()
            initial_pred_tube_mel = initial_pred_tube_mel[-1, :, :].detach().cpu().numpy().copy()
            initial_prod_tube_semvec = initial_prod_tube_semvec[-1, :].detach().cpu().numpy().copy()
            initial_pred_tube_semvec = initial_pred_tube_semvec[-1, :].detach().cpu().numpy().copy()

        else:
            initial_sig, initial_sr = speak(inv_normalize_cp(xx_new_numpy))
        
        initial_prod_mel = librosa_melspec(initial_sig, initial_sr)
        initial_prod_mel = normalize_mel_librosa(initial_prod_mel)
        initial_prod_mel.shape = initial_pred_mel.shape
        initial_prod_mel = torch.from_numpy(initial_prod_mel)
        initial_prod_mel = initial_prod_mel.to(self.device)

        # if past_cp prepend the first past_cp / 2 steps to target_mel
        if not past_cp is None:
            target_mel = torch.cat((initial_prod_mel[:, 0:(past_cp.shape[0] // 2), :], target_mel), dim=1)
            target_mel = target_mel.to(self.device)

        with torch.no_grad():
            initial_prod_semvec = self.embedder(initial_prod_mel, (torch.tensor(initial_prod_mel.shape[1]),))

        initial_prod_mel = initial_prod_mel[-1, :, :].detach().cpu().numpy().copy()
        initial_pred_mel = initial_pred_mel[-1, :, :].detach().cpu().numpy().copy()
        initial_prod_semvec = initial_prod_semvec[-1, :].detach().cpu().numpy().copy()
        initial_pred_semvec = initial_pred_semvec[-1, :].detach().cpu().numpy().copy()


        self.best_synthesis_acoustic = BestSynthesisAcoustic(np.Inf, initial_cp, initial_sig, initial_prod_mel, initial_pred_mel)  
        self.best_synthesis_semantic = BestSynthesisSemantic(np.Inf, initial_cp, initial_sig, initial_prod_semvec, initial_pred_semvec)
        if self.use_somatosensory_feedback:
            self.best_synthesis_somatosensory = BestSynthesisSomatosensory(np.Inf, np.Inf, np.Inf, initial_cp, initial_sig,
                                                                           initial_prod_tube,
                                                                           initial_pred_tube,
                                                                           initial_prod_tube_mel,
                                                                           initial_pred_tube_mel,
                                                                           initial_prod_tube_semvec,
                                                                           initial_pred_tube_semvec)

        # continue learning
        start_time = time.time()
        for ii_outer in tqdm(range(n_outer)):
            # imagine and plan
            pred_mel_steps_ii = list()
            prod_mel_steps_ii = list()
            cp_steps_ii = list()
            pred_semvec_steps_ii = list()
            prod_semvec_steps_ii = list()

            if self.use_somatosensory_feedback:
                pred_tube_steps_ii = list()
                prod_tube_steps_ii = list()
                pred_tube_mel_steps_ii = list()
                prod_tube_mel_steps_ii = list()
                pred_tube_semvec_steps_ii = list()
                prod_tube_semvec_steps_ii = list()

            for ii in range(n_inner):
                optimizer.zero_grad()

                pred_mel = self.pred_model(xx_new)
                if self.use_somatosensory_feedback:
                    pred_tube = self.cp_tube_model(xx_new)
                    pred_tube.retain_grad()
                    pred_tube_mel = self.tube_mel_model(pred_tube)

                if objective in ('semvec', 'acoustic_semvec'):
                    seq_length = pred_mel.shape[1]
                    self.embedder = self.embedder.train()
                    pred_semvec = self.embedder(pred_mel, (torch.tensor(seq_length),))
                    pred_semvec_steps_ii.append(pred_semvec[-1, :].detach().cpu().numpy().copy())

                    if self.use_somatosensory_feedback:
                        tube_seq_length = pred_tube.shape[1]
                        self.tube_embedder = self.tube_embedder.train()
                        pred_tube_semvec = self.tube_embedder(pred_tube, (torch.tensor(tube_seq_length),))
                        pred_tube_semvec_steps_ii.append(pred_tube_semvec[-1, :].detach().cpu().numpy().copy())

                # discrepancy,mel_loss, vel_loss, jerk_loss, pred_mel = constrained_criterion(tanh_straight_through(xx_new), target_mel)
                if objective == 'acoustic':
                    if self.use_somatosensory_feedback:
                        discrepancy, mel_loss, vel_loss, jerk_loss, tube_mel_loss = criterion(pred_mel, target_mel, xx_new, pred_tube_mel)
                    else:
                        discrepancy, mel_loss, vel_loss, jerk_loss = criterion(pred_mel, target_mel, xx_new)

                    if (ii + 1) % log_ii == 0:
                        planned_loss_steps.append(float(discrepancy.item()))
                        planned_mel_loss_steps.append(float(mel_loss.item()))
                        vel_loss_steps.append(float(vel_loss.item()))
                        jerk_loss_steps.append(float(jerk_loss.item()))
                        if self.use_somatosensory_feedback:
                            pred_tube_mel_loss_steps.append(float(tube_mel_loss.item()))

                        if log_semantics:
                            seq_length = pred_mel.shape[1]
                            pred_semvec = self.embedder(pred_mel, (torch.tensor(seq_length),))
                            pred_semvec_steps_ii.append(pred_semvec[-1, :].detach().cpu().numpy().copy())
                            semvec_loss = float(rmse_loss(pred_semvec, target_semvec).item())
                            pred_semvec_loss_steps.append(semvec_loss)
                            if self.use_somatosensory_feedback:
                                tube_seq_length = pred_tube.shape[1]
                                pred_tube_semvec = self.tube_embedder(pred_tube, (torch.tensor(tube_seq_length),))
                                pred_tube_semvec_steps_ii.append(pred_tube_semvec[-1, :].detach().cpu().numpy().copy())
                                tube_semvec_loss = 10*float(rmse_loss(pred_tube_semvec, target_semvec).item())
                                pred_tube_semvec_loss_steps.append(tube_semvec_loss)

                    if verbose:
                        print("Iteration %d" % ii)
                        print("Planned Loss: ", float(discrepancy.item()))
                        print("Mel Loss: ", float(mel_loss.item()))
                        print("Vel Loss: ", float(vel_loss.item()))
                        print("Jerk Loss: ", float(jerk_loss.item()))
                        if self.use_somatosensory_feedback:
                            print("Tube Mel Loss: ", float(tube_mel_loss.item()))
                        if log_semantics:
                            print("Semvec Loss: ", float(semvec_loss))
                            if self.use_somatosensory_feedback:
                                print("Tube Semvec Loss: ", float(tube_semvec_loss))

                elif objective == 'acoustic_semvec':
                    if self.use_somatosensory_feedback:
                        discrepancy, mel_loss, vel_loss, jerk_loss, semvec_loss, tube_mel_loss, tube_semvec_loss = criterion(pred_mel, target_mel,
                                                                                            pred_semvec, target_semvec,
                                                                                            xx_new, pred_tube_mel, pred_tube_semvec)
                    else:
                        discrepancy, mel_loss, vel_loss, jerk_loss, semvec_loss = criterion(pred_mel, target_mel,
                                                                                            pred_semvec, target_semvec,
                                                                                            xx_new)
                    if (ii + 1) % log_ii == 0:
                        planned_loss_steps.append(float(discrepancy.item()))
                        planned_mel_loss_steps.append(float(mel_loss.item()))
                        vel_loss_steps.append(float(vel_loss.item()))
                        jerk_loss_steps.append(float(jerk_loss.item()))
                        pred_semvec_loss_steps.append(float(semvec_loss.item()))
                        if self.use_somatosensory_feedback:
                            pred_tube_mel_loss_steps.append(float(tube_mel_loss.item()))
                            pred_tube_semvec_loss_steps.append(float(tube_semvec_loss.item()))

                    if verbose:
                        print("Iteration %d" % ii)
                        print("Planned Loss: ", float(discrepancy.item()))
                        print("Mel Loss: ", float(mel_loss.item()))
                        print("Vel Loss: ", float(vel_loss.item()))
                        print("Jerk Loss: ", float(jerk_loss.item()))
                        print("Semvec Loss: ", float(semvec_loss.item()))
                        if self.use_somatosensory_feedback:
                            print("Tube Mel Loss: ", float(tube_mel_loss.item()))
                            print("Tube Semvec Loss: ", float(tube_semvec_loss.item()))

                elif objective == 'semvec':
                    if self.use_somatosensory_feedback:
                        discrepancy, vel_loss, jerk_loss, semvec_loss, tube_semvec_loss = criterion(pred_semvec, target_semvec, xx_new, pred_tube_semvec)
                        mel_loss = rmse_loss(pred_mel, target_mel)
                        tube_mel_loss = rmse_loss(pred_tube_mel, target_mel)
                    else:
                        discrepancy, vel_loss, jerk_loss, semvec_loss = criterion(pred_semvec, target_semvec, xx_new)
                        mel_loss = rmse_loss(pred_mel, target_mel)

                    if (ii + 1) % log_ii == 0:
                        planned_loss_steps.append(float(discrepancy.item()))
                        vel_loss_steps.append(float(vel_loss.item()))
                        jerk_loss_steps.append(float(jerk_loss.item()))
                        pred_semvec_loss_steps.append(float(semvec_loss.item()))
                        planned_mel_loss_steps.append(float(mel_loss.item()))

                        if self.use_somatosensory_feedback:
                            pred_tube_semvec_loss_steps.append(float(tube_semvec_loss.item()))
                            pred_tube_mel_loss_steps.append(float(tube_mel_loss.item()))

                    if verbose:
                        print("Iteration %d" % ii)
                        print("Planned Loss: ", float(discrepancy.item()))
                        print("Mel Loss: ", float(mel_loss.item()))
                        print("Vel Loss: ", float(vel_loss.item()))
                        print("Jerk Loss: ", float(jerk_loss.item()))
                        print("Semvec Loss: ", float(semvec_loss.item()))
                        if self.use_somatosensory_feedback:
                            print("Tube Mel Loss: ", float(tube_mel_loss.item()))
                            print("Tube Semvec Loss: ", float(tube_semvec_loss.item()))

                else:
                    raise ValueError(f'unkown objective {objective}')

                discrepancy.backward()

                # if verbose:
                # print(f"grad.abs().max() {grad.abs().max()}")
                if xx_new.grad.max() > 10:
                    if verbose:
                        print("WARNING: gradient is larger than 10")
                if xx_new.grad.min() < -10:
                    if verbose:
                        print("WARNING: gradient is smaller than -10")

                if log_gradients:
                    grad_steps.append(xx_new.grad.detach().clone())

                if (ii + 1) % log_ii == 0:
                    xx_new_numpy = xx_new[-1, :, :].detach().cpu().numpy().copy()
                    cp_steps_ii.append(xx_new_numpy)

                    if self.use_somatosensory_feedback:
                        sig, sr, tube_info = speak_and_extract_tube_information(inv_normalize_cp(xx_new_numpy))
                        area_within_oral_cavity = get_area_info_within_oral_cavity(tube_info["tube_length_cm"], tube_info["tube_area_cm2"])
                        prod_tube = np.concatenate([area_within_oral_cavity,
                                                np.expand_dims(tube_info["incisor_pos_cm"], axis=1),
                                                np.expand_dims(tube_info["tongue_tip_side_elevation"], axis=1),
                                                np.expand_dims(tube_info["velum_opening_cm2"],axis=1)], axis=1)
                        prod_tube = normalize_tube(prod_tube)
                        prod_tube_steps_ii.append(prod_tube.copy())
                        prod_tube.shape = pred_tube.shape
                        prod_tube = torch.from_numpy(prod_tube)
                        prod_tube = prod_tube.to(self.device)

                        with torch.no_grad():
                            pred_tube = self.cp_tube_model(xx_new)
                            prod_tube_mel = self.tube_mel_model(prod_tube)

                        pred_tube_steps_ii.append(pred_tube[-1, :, :].detach().cpu().numpy().copy())

                        prod_tube_loss = rmse_loss(pred_tube, prod_tube)
                        prod_tube_loss_steps.append(float(prod_tube_loss.item()))

                        prod_tube_mel_loss = rmse_loss(prod_tube_mel, target_mel)
                        prod_tube_mel_loss_steps.append(float(prod_tube_mel_loss.item()))

                        pred_tube_mel_steps_ii.append(pred_tube_mel[-1, :, :].detach().cpu().numpy().copy())
                        prod_tube_mel_steps_ii.append(prod_tube_mel[-1, :, :].detach().cpu().numpy().copy())


                    else:
                        sig, sr = speak(inv_normalize_cp(xx_new_numpy))

                    if log_signals:
                        sig_steps.append(sig)

                    prod_mel = librosa_melspec(sig, sr)
                    prod_mel = normalize_mel_librosa(prod_mel)
                    prod_mel_steps_ii.append(prod_mel.copy())

                    prod_mel.shape = pred_mel.shape
                    prod_mel = torch.from_numpy(prod_mel)
                    prod_mel = prod_mel.to(self.device)
                    pred_mel_steps_ii.append(pred_mel[-1, :, :].detach().cpu().numpy().copy())

                    prod_loss = rmse_loss(prod_mel, target_mel)
                    prod_loss_steps.append(float(prod_loss.item()))

                    if verbose:
                        print("Produced Mel Loss: ", float(prod_loss.item()))
                        if self.use_somatosensory_feedback:
                            print("Produced Tube Loss: ", float(prod_tube_loss.item()))

                    if objective in ('semvec', 'acoustic_semvec') or log_semantics:
                        self.embedder = self.embedder.eval()
                        prod_semvec = self.embedder(prod_mel, (torch.tensor(prod_mel.shape[1]),))
                        prod_semvec_steps_ii.append(prod_semvec[-1, :].detach().cpu().numpy().copy())

                        prod_semvec_loss = 10*rmse_loss(prod_semvec, target_semvec)
                        prod_semvec_loss_steps.append(float(prod_semvec_loss.item()))

                        if self.use_somatosensory_feedback:
                            self.tube_embedder = self.tube_embedder.eval()
                            prod_tube_semvec = self.tube_embedder(prod_tube, (torch.tensor(prod_tube.shape[1]),))
                            prod_tube_semvec_steps_ii.append(prod_tube_semvec[-1, :].detach().cpu().numpy().copy())

                            prod_tube_semvec_loss = 10*rmse_loss(prod_tube_semvec, target_semvec)
                            prod_tube_semvec_loss_steps.append(float(prod_tube_semvec_loss.item()))

                        if verbose:
                            print("Produced Semvec Loss: ", float(prod_semvec_loss.item()))
                            if self.use_somatosensory_feedback:
                                print("Produced Tube Semvec Loss: ", float(prod_tube_semvec_loss.item()))
                            print("")

                        new_synthesis_acoustic = BestSynthesisAcoustic(float(prod_loss.item()), xx_new_numpy, sig, prod_mel[-1, :, :].detach().cpu().numpy().copy(), pred_mel[-1, :, :].detach().cpu().numpy().copy())
                        new_synthesis_semantic = BestSynthesisSemantic(float(prod_semvec_loss.item()), xx_new_numpy, sig, prod_semvec[-1, :].detach().cpu().numpy().copy(), pred_semvec[-1, :].detach().cpu().numpy().copy())

                        if self.best_synthesis_acoustic.mel_loss > new_synthesis_acoustic.mel_loss:
                            self.best_synthesis_acoustic = new_synthesis_acoustic
                        if self.best_synthesis_semantic.semvec_loss > new_synthesis_semantic.semvec_loss:
                            self.best_synthesis_semantic = new_synthesis_semantic

                        if self.use_somatosensory_feedback:
                            new_synthesis_somatosensory = BestSynthesisSomatosensory(float(prod_tube_loss.item()), float(prod_tube_mel_loss.item()), float(prod_tube_semvec_loss.item()),
                                                                                           xx_new_numpy, sig,
                                                                                           prod_tube,
                                                                                           pred_tube[-1, :].detach().cpu().numpy().copy(),
                                                                                           prod_tube_mel[-1, :].detach().cpu().numpy().copy(),
                                                                                           pred_tube_mel[-1, :].detach().cpu().numpy().copy(),
                                                                                           prod_tube_semvec[-1, :].detach().cpu().numpy().copy(),
                                                                                           pred_tube_semvec[-1, :].detach().cpu().numpy().copy())
                            if self.best_synthesis_somatosensory.tube_loss > new_synthesis_somatosensory.tube_loss:
                                self.best_synthesis_somatosensory = new_synthesis_somatosensory


                    else:
                        new_synthesis_acoustic = BestSynthesisAcoustic(float(prod_loss.item()), xx_new_numpy, sig, prod_mel[-1, :, :].detach().cpu().numpy().copy(),pred_mel[-1, :, :].detach().cpu().numpy().copy())
                        if self.best_synthesis_acoustic.mel_loss > new_synthesis_acoustic.mel_loss:
                            self.best_synthesis_acoustic = new_synthesis_acoustic

                        if self.use_somatosensory_feedback:
                            new_synthesis_somatosensory = BestSynthesisSomatosensory(float(prod_tube_loss.item()), float(prod_tube_mel_loss.item()), np.Inf,
                                                                                           xx_new_numpy, sig,
                                                                                           prod_tube,
                                                                                           pred_tube[-1, :].detach().cpu().numpy().copy(),
                                                                                           prod_tube_mel[-1, :].detach().cpu().numpy().copy(),
                                                                                           pred_tube_mel[-1, :].detach().cpu().numpy().copy(),
                                                                                           None,
                                                                                           None)
                            if self.best_synthesis_somatosensory.tube_loss > new_synthesis_somatosensory.tube_loss:
                                self.best_synthesis_somatosensory = new_synthesis_somatosensory
                        
                    if verbose:
                        print("")

                optimizer.step()

                with torch.no_grad():
                    xx_new.data = xx_new.data.clamp(-1.05, 1.05) # clamp between -1.05 and 1.05
                    if self.smiling:
                        #Vocal tract parameters: "HX HY JX JA LP LD VS VO TCX TCY TTX TTY TBX TBY TRX TRY TS1 TS2 TS3"
                        #Glottis parameters: "f0 pressure x_bottom x_top chink_area lag rel_amp double_pulsing pulse_skewness flutter aspiration_strength "
                        # keep LP and HY at the maximum
                        xx_new.data[:, :, 4] = -1.0  # "LP"
                        xx_new.data[:, :, 1] = 1.0  # "HY"

                    if not past_cp is None:
                        xx_new.data[:, 0:past_cp_torch.shape[0], :] = past_cp_torch


            if plot:
                target_mel_ii = target_mel[-1, :, :].detach().cpu().numpy().copy()
                prod_mel_ii = prod_mel[-1, :, :].detach().cpu().numpy().copy()
                pred_mel_ii = pred_mel[-1, :, :].detach().cpu().numpy().copy()

                if plot is True:
                    visualize.plot_mels(True, target_mel_ii, initial_pred_mel,
                            initial_prod_mel, pred_mel_ii, prod_mel_ii)
                else:
                    visualize.plot_mels(f"{plot}_{ii_outer:03d}.png",
                            target_mel_ii, initial_pred_mel, initial_prod_mel,
                            pred_mel_ii, prod_mel_ii)

            prod_mel_steps.append(prod_mel_steps_ii)
            if log_cps:
                cp_steps.append(cp_steps_ii)
            pred_mel_steps.append(pred_mel_steps_ii)
            pred_semvec_steps.append(pred_semvec_steps_ii)
            prod_semvec_steps.append(prod_semvec_steps_ii)

            if self.use_somatosensory_feedback:
                prod_tube_steps.append(prod_tube_steps_ii)
                pred_tube_steps.append(pred_tube_steps_ii)
                prod_tube_mel_steps.append(prod_tube_mel_steps_ii)
                pred_tube_mel_steps.append(pred_tube_mel_steps_ii)
                pred_tube_semvec_steps.append(pred_tube_semvec_steps_ii)
                prod_tube_semvec_steps.append(prod_tube_semvec_steps_ii)


            # execute and continue learning
            if continue_learning:
                produced_data_ii = pd.DataFrame(columns=['vector', 'cp_norm', 'melspec_norm_synthesized', "tube_norm", 'segment_data'])
                produced_data_ii["cp_norm"] = cp_steps_ii
                produced_data_ii["melspec_norm_synthesized"] = prod_mel_steps_ii
                produced_data_ii["vector"] = [target_semvec[0].detach().cpu().numpy().copy() for _ in range(len(produced_data_ii))]
                produced_data_ii["segment_data"] = False
                if self.use_somatosensory_feedback:
                    produced_data_ii["tube_norm"] = prod_tube_steps_ii

                if add_training_data_pred or add_training_data_inv:
                    # update with new sample
                    if len(produced_data_ii) < int(0.5 * batch_size) * n_batches: # too few samples synthesized for filling 50% of batch_size * n_batches
                        batch_train_produced = random.sample(range(len(produced_data_ii)), k=len(produced_data_ii))
                        batch_train = random.sample(range(len(self.continue_data)), k=len(produced_data_ii))
                        n_train_batches = int(np.ceil(2*len(produced_data_ii)/batch_size))
                        reduced_last_batch = (2*len(produced_data_ii))%batch_size
                        print("Enhanced training data")
                        print(f"Not enough data produced to fill 50% of {n_batches} batches...")
                        if n_train_batches < n_batches:
                            print(f"Training on {n_train_batches} batches instead...")
                        if reduced_last_batch > 0:
                            print(f"Last batch reduced to {reduced_last_batch} samples instead of {batch_size}...")
                        print(" ")
                    else:
                        batch_train = random.sample(range(len(self.continue_data)), k=int(0.5 * batch_size) * n_batches)
                        batch_train_produced = random.sample(range(len(produced_data_ii)), k=int(0.5 * batch_size) * n_batches)
                        n_train_batches = n_batches

                    if self.use_somatosensory_feedback:
                        train_data_samples = self.continue_data[['vector','cp_norm', 'melspec_norm_synthesized', 'tube_norm','segment_data']].iloc[
                            batch_train].reset_index(drop=True)
                    else:
                        train_data_samples = self.continue_data[['vector', 'cp_norm', 'melspec_norm_synthesized', 'segment_data']].iloc[
                            batch_train].reset_index(drop=True)

                    produced_data_ii_samples = produced_data_ii.iloc[batch_train_produced].reset_index(drop=True)
                    continue_data_ii = pd.concat([train_data_samples, produced_data_ii_samples])


                    continue_data_ii["lens_input"] = np.array(continue_data_ii["cp_norm"].apply(len), dtype=int)
                    continue_data_ii["lens_output"] = np.array(continue_data_ii["melspec_norm_synthesized"].apply(len),dtype=int)
                    continue_data_ii.sort_values(by="lens_input", inplace=True)

                    del train_data_samples

                if len(produced_data_ii) <  batch_size * n_batches: # too few samples synthesized for filling batch_size * n_batches
                    batch_train_produced = random.sample(range(len(produced_data_ii)), k=len(produced_data_ii))
                    n_train_batches = int(np.ceil(len(produced_data_ii)/batch_size))
                    reduced_last_batch = len(produced_data_ii)%batch_size
                    print("Produced training data")
                    print(f"Not enough data produced to fill {n_batches} batches...")
                    if n_train_batches < n_batches:
                        print(f"Training on {n_train_batches} batches instead...")
                    if reduced_last_batch > 0:
                        print(f"Last batch reduced to {reduced_last_batch} samples instead of {batch_size}...")
                    print(" ")
                else:
                    batch_train_produced = random.sample(range(len(produced_data_ii)), k= batch_size * n_batches)
                    n_train_batches = n_batches

                produced_data_ii_samples = produced_data_ii.iloc[batch_train_produced].reset_index(drop=True)
                produced_data_ii_samples["lens_input"] = np.array(produced_data_ii_samples["cp_norm"].apply(len), dtype=int)
                produced_data_ii_samples["lens_output"] = np.array(produced_data_ii_samples["melspec_norm_synthesized"].apply(len),
                                                           dtype=int)
                produced_data_ii_samples.sort_values(by="lens_input", inplace=True)

                if add_training_data_pred:
                    training_data_pred = continue_data_ii
                else:
                    training_data_pred = produced_data_ii_samples

                training_length_dict_pred = {}
                lens_training_cps_pred = np.asarray(training_data_pred.lens_input)
                lengths, counts = np.unique(lens_training_cps_pred, return_counts=True)
                sorted_training_length_keys_pred = np.sort(lengths)
                for length in sorted_training_length_keys_pred:
                    training_length_dict_pred[length] = np.where(lens_training_cps_pred == length)[0]

                inps = training_data_pred["cp_norm"]
                tgts = training_data_pred["melspec_norm_synthesized"]

                lens_input = torch.tensor(np.array(training_data_pred.lens_input)).to(self.device)
                lens_output = torch.tensor(np.array(training_data_pred.lens_output)).to(self.device)

                if self.use_somatosensory_feedback:
                    tgts_tube = training_data_pred["tube_norm"]
                    lens_output_tube = torch.tensor(np.array(training_data_pred.lens_input)).to(self.device)

                if continue_learning_inv:
                    if add_training_data_inv:
                        training_data_inv = continue_data_ii
                    else:
                        training_data_inv = produced_data_ii_samples

                    training_length_dict_inv = {}
                    lens_training_cps_inv = np.asarray(training_data_inv.lens_input)
                    lengths, counts = np.unique(lens_training_cps_inv, return_counts=True)
                    sorted_training_length_keys_inv = np.sort(lengths)
                    for length in sorted_training_length_keys_inv:
                        training_length_dict_inv[length] = np.where(lens_training_cps_inv == length)[0]

                    inps_inv = training_data_inv["melspec_norm_synthesized"]
                    tgts_inv = training_data_inv["cp_norm"]

                    lens_input_inv = torch.tensor(np.array(training_data_inv.lens_output)).to(self.device)
                    lens_output_inv = torch.tensor(np.array(training_data_inv.lens_input)).to(self.device)



                for e in range(n_epochs):
                    epoch_ii = self.create_epoch_batches(training_data_pred, batch_size, shuffle=True,
                                                         same_size_batching=True,
                                                         training_length_dict=training_length_dict_pred)
                    avg_loss = list()
                    if continue_learning_tube and self.use_somatosensory_feedback:
                        avg_loss_tube = list()
                        avg_loss_tube_mel = list()

                    for j in epoch_ii:
                        lens_input_j = lens_input[j]
                        batch_input = inps.iloc[j]
                        batch_input = pad_batch_online(lens_input_j, batch_input, self.device)

                        lens_output_j = lens_output[j]
                        batch_output = tgts.iloc[j]
                        batch_output = pad_batch_online(lens_output_j, batch_output, self.device)


                        Y_hat = self.pred_model(batch_input, lens_input_j)

                        self.pred_optimizer.zero_grad()
                        pred_loss = self.pred_criterion(Y_hat, batch_output)
                        pred_loss.backward()
                        self.pred_optimizer.step()

                        avg_loss.append(float(pred_loss.item()))

                        if continue_learning_tube and self.use_somatosensory_feedback:
                            lens_output_tube_j = lens_output_tube[j]
                            batch_output_tube = tgts_tube.iloc[j]
                            batch_output_tube = pad_batch_online(lens_output_tube_j, batch_output_tube, self.device)

                            # learn cp_tube_model
                            Y_hat = self.cp_tube_model(batch_input, lens_input_j)

                            self.tube_optimizer.zero_grad()
                            tube_loss = self.tube_criterion(Y_hat, batch_output_tube)
                            tube_loss.backward()
                            self.tube_optimizer.step()

                            avg_loss_tube.append(float(tube_loss.item()))

                            # learn tube_mel_model
                            Y_hat = self.tube_mel_model(batch_output_tube)

                            self.tube_mel_optimizer.zero_grad()
                            tube_mel_loss = self.tube_mel_criterion(Y_hat, batch_output)
                            tube_mel_loss.backward()
                            self.tube_mel_optimizer.step()

                            avg_loss_tube_mel.append(float(tube_mel_loss.item()))

                    pred_model_loss.append(np.mean(avg_loss))

                    if continue_learning_tube and self.use_somatosensory_feedback:
                        tube_model_loss.append(np.mean(avg_loss_tube))
                        tube_mel_model_loss.append(np.mean(avg_loss_tube_mel))

                if continue_learning_inv:
                    for e in range(n_epochs):
                        epoch_ii = self.create_epoch_batches(training_data_inv, batch_size, shuffle=True,
                                                             same_size_batching=True,
                                                             training_length_dict=training_length_dict_inv)
                        avg_loss_inv = list()
                        for j in epoch_ii:
                            lens_input_j = lens_input_inv[j]
                            batch_input = inps_inv.iloc[j]
                            batch_input = pad_batch_online(lens_input_j, batch_input, self.device)

                            lens_output_j = lens_output_inv[j]
                            batch_output = tgts_inv.iloc[j]
                            batch_output = pad_batch_online(lens_output_j, batch_output, self.device)


                            Y_hat = self.inv_model(batch_input, lens_input_j)

                            self.inv_optimizer.zero_grad()
                            inv_loss = self.inv_criterion(Y_hat, batch_output)
                            inv_sub_losses = inv_loss[1:]  # sublosses
                            inv_loss = inv_loss[0]  # total loss
                            inv_loss.backward()
                            self.inv_optimizer.step()

                            avg_loss_inv.append(float(inv_loss.item()))

                        inv_model_loss.append(np.mean(avg_loss_inv))


                if not self.continue_data is None:
                    self.continue_data = pd.concat([self.continue_data, produced_data_ii]).reset_index(drop=True)
                    if len(self.continue_data) > self.continue_data_limit:
                        random_sample = random.sample(range(len(self.continue_data)), k=self.continue_data_limit)
                        self.continue_data = self.continue_data.iloc[random_sample].reset_index(drop=True)

                del produced_data_ii_samples
                del produced_data_ii
                del training_data_pred
                if continue_learning_inv:
                    del training_data_inv
                if add_training_data_pred or add_training_data_inv:
                    del continue_data_ii

        planned_cp = xx_new[-1, :, :].detach().cpu().numpy()
        prod_sig = sig
        prod_sr = sr

        with torch.no_grad():
            pred_mel = self.pred_model(xx_new)
            self.embedder = self.embedder.eval()
            pred_semvec = self.embedder(pred_mel, (torch.tensor(pred_mel.shape[1]),))
            prod_semvec = self.embedder(prod_mel, (torch.tensor(prod_mel.shape[1]),))

        target_mel = target_mel[-1, :, :].detach().cpu().numpy()
        prod_mel = prod_mel[-1, :, :].detach().cpu().numpy()
        pred_mel = pred_mel[-1, :, :].detach().cpu().numpy()
        prod_semvec = prod_semvec[-1, :].detach().cpu().numpy()
        pred_semvec = pred_semvec[-1, :].detach().cpu().numpy()

        if self.use_somatosensory_feedback:
            with torch.no_grad():
                pred_tube = self.cp_tube_model(xx_new)
                prod_tube_mel = self.tube_mel_model(prod_tube)
                pred_tube_mel = self.tube_mel_model(pred_tube)

                self.tube_embedder = self.tube_embedder.eval()
                prod_tube_semvec = self.tube_embedder(prod_tube, (torch.tensor(prod_tube.shape[1]),))
                pred_tube_semvec = self.tube_embedder(pred_tube, (torch.tensor(prod_tube.shape[1]),))

            prod_tube = prod_tube[-1, :, :].detach().cpu().numpy()
            pred_tube = pred_tube[-1, :, :].detach().cpu().numpy()
            prod_tube_mel = prod_tube_mel[-1, :, :].detach().cpu().numpy()
            pred_tube_mel = pred_tube_mel[-1, :, :].detach().cpu().numpy()
            prod_tube_semvec = prod_tube_semvec[-1, :].detach().cpu().numpy()
            pred_tube_semvec = pred_tube_semvec[-1, :].detach().cpu().numpy()


        print("--- %.2f min ---" % ((time.time() - start_time) / 60))

        #  0. planned_cp
        #  1. initial_cp
        #  3. initial_sig
        #  4. initial_sr
        #  5. initial_prod_mel
        #  6. initial_pred_mel
        #  7. target_sig
        #  8. target_sr
        #  9. target_mel
        # 10. prod_sig
        # 11. prod_sr
        # 12. prod_mel
        # 13. pred_mel
        # 14. initial_prod_semvec
        # 15. initial_pred_semvec
        # 16. prod_semvec
        # 17. pred_semvec
        # 18. prod_loss_steps
        # 19. planned_loss_steps
        # 20. planned_mel_loss_steps
        # 21. vel_loss_steps
        # 22. jerk_loss_steps
        # 23. pred_semvec_loss_steps
        # 24. prod_semvec_loss_steps
        # 25. cp_steps
        # 26. pred_semvec_steps
        # 27. prod_semvec_steps
        # 28. grad_steps
        # 29. sig_steps
        # 30. prod_mel_steps
        # 31. pred_mel_steps
        # 32. pred_model_loss
        # 33. inv_model_loss

        if self.use_somatosensory_feedback:
            return PlanningResultsWithSomatosensory(planned_cp, initial_cp, initial_sig, initial_sr, initial_prod_mel, initial_pred_mel, initial_prod_tube, initial_pred_tube, initial_prod_tube_mel, initial_pred_tube_mel,
                                   target_sig, target_sr, target_mel, prod_sig, prod_sr, prod_mel,
                                   pred_mel, prod_tube, pred_tube, prod_tube_mel, pred_tube_mel, initial_prod_semvec, initial_pred_semvec, initial_prod_tube_semvec, initial_pred_tube_semvec, prod_semvec, pred_semvec, prod_tube_semvec, pred_tube_semvec,
                                   prod_loss_steps, planned_loss_steps,
                                   planned_mel_loss_steps, vel_loss_steps, jerk_loss_steps,
                                   pred_semvec_loss_steps, prod_semvec_loss_steps, prod_tube_loss_steps, pred_tube_mel_loss_steps, prod_tube_mel_loss_steps, pred_tube_semvec_loss_steps, prod_tube_semvec_loss_steps, cp_steps,
                                   pred_semvec_steps, prod_semvec_steps, grad_steps, sig_steps,
                                   prod_mel_steps, pred_mel_steps, prod_tube_steps, pred_tube_steps, prod_tube_mel_steps, pred_tube_mel_steps, prod_tube_semvec_steps, pred_tube_semvec_steps, pred_model_loss, inv_model_loss, tube_model_loss, tube_mel_model_loss)

        else:
            return PlanningResults(planned_cp, initial_cp, initial_sig, initial_sr, initial_prod_mel, initial_pred_mel,
                    target_sig, target_sr, target_mel, prod_sig, prod_sr, prod_mel,
                    pred_mel, initial_prod_semvec, initial_pred_semvec, prod_semvec, pred_semvec, prod_loss_steps, planned_loss_steps,
                    planned_mel_loss_steps, vel_loss_steps, jerk_loss_steps,
                    pred_semvec_loss_steps, prod_semvec_loss_steps, cp_steps,
                    pred_semvec_steps, prod_semvec_steps, grad_steps, sig_steps,
                    prod_mel_steps, pred_mel_steps, pred_model_loss, inv_model_loss)




"""
This module contains useful functions to visualize and safe results.

"""

import os

import soundfile as sf
from matplotlib import pyplot as plt
import librosa.display
from matplotlib import cm

from . import util


def visualize_results(results, condition='prefix', folder='data'):
    """
    Stores all results in data/ folder.

    """
    base_name = os.path.join(folder, f'{condition}')

    # save mel plot
    plot_mels(f"{base_name}_mel.png", results.target_mel, results.initial_pred_mel,
            results.initial_prod_mel, results.pred_mel, results.prod_mel)

    # save audio
    target_sr = prod_sr = 44100
    sf.write(f'{base_name}_planned.flac', results.prod_sig, results.prod_sr)
    sf.write(f'{base_name}_initial.flac', results.initial_sig, results.initial_sr)
    sf.write(f'{base_name}_target.flac', results.target_sig, results.target_sr)

    # save loss plot
    fig, ax = plt.subplots(figsize=(15, 8), facecolor="white")
    ax.plot(results.planned_loss_steps, label="planned loss", c="C0")
    ax.legend()
    fig.tight_layout()
    fig.savefig(f"{base_name}_loss.png")

    fig, ax = plt.subplots(figsize=(15, 8), facecolor="white")
    ax.plot(results.prod_loss_steps, label="produced mel loss", c="C1")
    ax.plot(results.planned_mel_loss_steps, label="planned mel loss", c="C0")
    ax.legend()
    fig.tight_layout()
    fig.savefig(f"{base_name}_loss_mel.png")

    # save subloss plot
    fig, ax = plt.subplots(figsize=(15, 8), facecolor="white")
    ax.plot(results.vel_loss_steps, label="vel loss", c="C2")
    ax.plot(results.jerk_loss_steps, label="jerk loss", c="C3")
    ax.legend()
    fig.tight_layout()
    fig.savefig(f"{base_name}_loss_subloss.png")

    # save semvec loss plot
    fig, ax = plt.subplots(figsize=(15, 8), facecolor="white")
    ax.plot(results.pred_semvec_loss_steps, label="planned semvec loss", c="C0")
    ax.plot(results.prod_semvec_loss_steps, label="produced semvec loss", c="C1")
    ax.legend()
    fig.tight_layout()
    fig.savefig(f"{base_name}_loss_semvec.png")

    # save cp change plot
    fig = plt.figure(figsize=(15, 12))
    ax1 = fig.add_axes([0.1, 0.68, 0.88, 0.30], xticklabels=[])
    ax2 = fig.add_axes([0.1, 0.36, 0.88, 0.30], xticklabels=[], sharex=ax1)
    ax3 = fig.add_axes([0.1, 0.04, 0.88, 0.30], xticklabels=[], sharex=ax1)
    img1 = results.initial_cp  #target['cp'].iloc[0]
    img2 = results.planned_cp
    img3 = img2 - img1
    ax1.plot(img1[:, 8:16])  # , label='tongue'
    ax1.plot(img1[:, 19:20], label='f0')
    ax1.set_ylabel("initial")
    ax2.plot(img2[:, 8:16], label='tongue')
    ax2.plot(img2[:, 19:20], label='f0')
    ax2.set_ylabel("optimized")
    ax3.plot(img3[:, 8:16], label='tongue')
    ax3.plot(img3[:, 19:20], label='f0')
    ax3.set_ylabel("difference")
    ax1.legend()
    fig.tight_layout()
    fig.savefig(f'{base_name}_cps.png')

    # save svgs and create mp4s
    path = f"{base_name}_initial_svgs/"
    if not os.path.exists(path):
        os.mkdir(path)
    util.export_svgs(util.inv_normalize_cp(results.initial_cp), path=path)
    system_call = f'cd {path}; /usr/bin/ffmpeg -hide_banner -loglevel error -r 80 -width 768 -i tract%05d.svg -i ../{condition}_initial.flac -c:v libx264 -pix_fmt yuv420p ../{condition}_initial_80Hz.mp4'
    return_value = os.system(system_call)
    if return_value != 0:
        print("WARNING: creating the initial animation went wrong")
    ## recode to 60 Hz
    #system_call = f'cd {path}; /usr/bin/ffmpeg -loglevel error -i ../{condition}_initial_80Hz.mp4 -r 60 ../{condition}_initial_60Hz.mp4'
    #return_value = os.system(system_call)
    #if return_value != 0:
    #    print("WARNING: creating the initial animation went wrong")

    path = f"{base_name}_planned_svgs/"
    if not os.path.exists(path):
        os.mkdir(path)
    util.export_svgs(util.inv_normalize_cp(results.planned_cp), path=path)
    system_call = f'cd {path}; /usr/bin/ffmpeg -hide_banner -loglevel error -r 80 -width 768 -i tract%05d.svg -i ../{condition}_planned.flac -c:v libx264 -pix_fmt yuv420p ../{condition}_planned_80Hz.mp4'
    return_value = os.system(system_call)
    if return_value != 0:
        print("WARNING: creating the planning animation went wrong")
    ## recode to 60 Hz
    #system_call = f'cd {path}; /usr/bin/ffmpeg -loglevel error -i ../{condition}_planned_80Hz.mp4 -r 60 ../{condition}_planned_60Hz.mp4'
    #return_value = os.system(system_call)
    #if return_value != 0:
    #    print("WARNING: creating the planned animation went wrong")

    plt.show()  # this shows all saved figures


def plot_mels(file_name, target_mel, initial_pred_mel, initial_prod_mel,
        pred_mel, prod_mel):
    """
    Plots target, initial prediction, initial production, prediction and
    production log mel spectrograms.

    Parameters
    ==========
    file_name : str or True
    target_mel : np.array
    initial_pred_mel : np.array
    initial_prod_mel : np.array
    pred_mel : np.array
    prod_mel : np.array

    """
    fig, ax = plt.subplots(nrows=6, figsize=(15, 18), facecolor="white")
    librosa.display.specshow(target_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[0],
            cmap=cm.magma)
    ax[0].set_title("Target", fontsize=18)
    librosa.display.specshow(initial_prod_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[1],
            cmap=cm.magma)
    ax[1].set_title("Initial Produced", fontsize=18)
    librosa.display.specshow(initial_pred_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[2],
            cmap=cm.magma)
    ax[2].set_title("Initial Prediction", fontsize=18)
    librosa.display.specshow(pred_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[3],
            cmap=cm.magma)
    ax[3].set_title("Planned Prediction", fontsize=18)
    librosa.display.specshow(prod_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[4],
            cmap=cm.magma)
    ax[4].set_title("Planned Produced", fontsize=18)
    librosa.display.specshow(target_mel.T, y_axis='mel',
            x_axis='time', sr=44100, hop_length=220, ax=ax[5],
            cmap=cm.magma)
    ax[5].set_title("Target", fontsize=18)

    ax[0].set_xticks([])
    ax[1].set_xticks([])
    ax[2].set_xticks([])
    ax[3].set_xticks([])
    ax[4].set_xticks([])
    ax[0].set_xlabel("")
    ax[1].set_xlabel("")
    ax[2].set_xlabel("")
    ax[3].set_xlabel("")
    ax[4].set_xlabel("")
    ax[5].set_xlabel("Time (s)", fontsize=15)
    ax[0].set_ylabel("Hz", fontsize=15)
    ax[1].set_ylabel("Hz", fontsize=15)
    ax[2].set_ylabel("Hz", fontsize=15)
    ax[3].set_ylabel("Hz", fontsize=15)
    ax[4].set_ylabel("Hz", fontsize=15)
    ax[5].set_ylabel("Hz", fontsize=15)

    fig.tight_layout()

    if file_name is True:  # only if identical to "True" interactive
                           # blocking plotting
       plt.show()
    else:
        fig.savefig(file_name)




"""
paule - Predictive Articulatory speech synthesis Utilizing Lexical Embeddings
=============================================================================

*paule* is a prediction based control and copy synthesis model for the
articulatory speech synthesiser VocalTractLab (VTL; vocaltractlab.de). *paule* can
plan control parameter (cp) trajectories for the VTL synthesiser for a target
acoustics or target lexical or semantic embedding. To achieve this it uses
an inverse and and a predictive sequence-to-sequence model and a
sequence-to-vector classifier model. All models are
implemented in pytorch and pretrained weights are available.

"""

print("WARNING! The *paule* package is still in alpha. "
      "To download pretrained weights for the PAULE model, which is defined in "
      "`paule.paule.Paule`, you can invoke "
      "`paule.util.download_pretrained_weights()`.")

import os
import sys
import multiprocessing as mp
from pip._vendor import pkg_resources
try:
    from importlib.metadata import requires
except ModuleNotFoundError:  # python 3.7 and before
    requires = None
try:
    from packaging.requirements import Requirement
except ModuleNotFoundError:  # this should only happend during setup phase
    Requirement = None

try:
    from importlib import metadata
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    import toml
    __version__ = toml.load(os.path.join(os.path.dirname(__file__), "../pyproject.toml"))["tool"]["poetry"]["version"] + "dev"


def sysinfo():
    """
    Prints system the dependency information
    """
    if requires:
        dependencies = [Requirement(req).name for req in requires('paule')
                        if not Requirement(req).marker]

    header = ("Paule Information\n"
              "=================\n\n")

    general = ("General Information\n"
               "-------------------\n"
               "Python version: {}\n"
               "Paule version: {}\n\n").format(sys.version.split()[0], __version__)

    uname = platform.uname()
    osinfo = ("Operating System\n"
              "----------------\n"
              "OS: {s.system} {s.machine}\n"
              "Kernel: {s.release}\n"
              "CPU: {cpu_count}\n").format(s=uname, cpu_count=mp.cpu_count())

    if uname.system == "Linux":
        _, *lines = os.popen("free -m").readlines()
        for identifier in ("Mem:", "Swap:"):
            memory = [line for line in lines if identifier in line]
            if len(memory) > 0:
                _, total, used, *_ = memory[0].split()
            else:
                total, used = '?', '?'
            osinfo += "{} {}MiB/{}MiB\n".format(identifier, used, total)

    osinfo += "\n"

    deps = ("Dependencies\n"
            "------------\n")

    if requires:
        deps += "\n".join("{pkg.__name__}: {pkg.__version__}".format(pkg=__import__(dep))
                          for dep in dependencies)
    else:
        deps = 'You need Python 3.8 or higher to show dependencies.'

    print(header + general + osinfo + deps)

